r"""
Transient heat equation with time-dependent source term, three different
material domains and Newton type boundary condition loss term.

Description
-----------

This example is inspired by the Laser Powder Bed Fusion additive manufacturing
process. A laser source deposits a flux on a circular surface (a full layer of
the cylinder being built) at regular time intervals. The heat propagates into
the build plate and to the sides of the cylinder into powder which is
relatively bad at conductiong heat. Heat losses through Newton type heat
exchange occur both at the surface where the heat is being deposited and at the
bottom plate.

The PDE for this physical process implies to find :math:`T(x, t)` for :math:`x
\in \Omega, t \in [0, t_{\rm final}]` such that:

.. math::
    \left\lbrace
    \begin{aligned}
    \rho(x)c_p(x)\frac{\partial T}{\partial t}(x, t)
    = -\nabla \cdot \left( -k(x) \underline{\nabla} T(x, t) \right) &&
    \forall x \in \Omega, \forall t \in [0, T_{max}] \\
    -k \underline{\nabla} T(x, t)\cdot \underline{n}
    = q(t)+h(T-T_\infty) && \forall x \in \Gamma_{source} \\
    -k \underline{\nabla} T(x, t)\cdot \underline{n}
    = q(t)+h(T-T_\infty) && \forall x \in \Gamma_{plate} \\
    \underline{\nabla} T(x, t)\cdot \underline{n}
    = 0 && \forall x \in \Gamma \setminus(\Gamma_{source}
    \cap\Gamma_{plate})
    \end{aligned}
    \right.

The weak formulation solved using `sfepy` is to find a discretized field
:math:`T` that satisfies:

.. math::
    \begin{aligned}
    \int_\Omega\rho c_p \frac{\partial T}{\partial t}(x, t) \, s
    + \int_\Omega \underline{\nabla} s \cdot (k(x)
    \underline{\nabla}
    T) =  \int_\Gamma -k \underline{\nabla} T \cdot \underline{n} s \\
    = \int_{\Gamma_{source}} q(t)   +
    \int_{\Gamma_{source}} h(T-T_\infty)
    + \int_{\Gamma_{plate}} h(T-T_\infty)  && \forall s
    \end{aligned}

Uniform initial conditions are used as a starting point of the simulation.

Usage examples
--------------
The script can be run with::

    sfepy-run sfepy/examples/diffusion/time_heat_equation_multi_material.py

The last time-step result field can then be visualized as isosurfaces with::

    sfepy-view multi_material_cylinder_plate.119.vtk -i 10 -l

The resulting time evolution of temperatures is saved as an image file in the
output directory (heat_probe_time_evolution.png).

This script uses SI units (meters, kilograms, Joules...) except for
temperature, which is expressed in degrees Celsius.
"""
import numpy as nm
from sfepy import data_dir
from sfepy.discrete.probes import LineProbe
import matplotlib.pyplot as plt
import os

nominal_heat_flux = 6.36e5
alpha = 0.25
t_start = nm.array([0., 20., 40.])  # times when heating starts (seconds)
t_stop = nm.array([10., 30., 50.])  # times when heating stops (seconds)

T0 = 25.  # °C
h = 50.  # W/m2/K
mm = 1e-3

filename_mesh = data_dir + '/meshes/3d/multi_material_cylinder_plate.vtk'

materials = {
    'powder': ({'lam': 0.16,
                'rho_cp': 1330. * 650.},),
    'cylinder': ({'lam': 153.,
                  'rho_cp': 2760. * 935.,
                  'lam_vec': 173. * nm.eye(3)},),
    'plate': ({'lam': 153.,
               'rho_cp': 2660. * 927.,
               'lam_vec': 153. * nm.eye(3)},),

    'heat_flux_defined_by_func': 'get_flux_value',
    'heat_loss': ({'h_bot': -h, 'T_bot_inf': T0,
                   'h_top': -h, 'T_top_inf': T0},)
}

regions = {
    'Omega': 'all',
    'Omega_Cylinder': 'cells of group 4',
    'Omega_Powder': 'cells of group 3',
    'Omega_Plate': 'cells of group 1 +v cells of group 2',
    'Gamma_Plate': ('vertices in (z < -9.95e-3)', 'facet'),
    'Gamma_Source': ('vertices of surface *v r.Omega_Cylinder', 'face'),
}


def get_flux_value(ts, coors, mode=None, **kwargs):
    """Defines heat flux as a function of time."""
    if mode == 'qp':
        shape = (coors.shape[0], 1, 1)
        if nm.any((ts.time >= t_start) & (ts.time <= t_stop)):
            flux = alpha * nominal_heat_flux
        else:
            flux = 0.
        val = flux * nm.ones(shape, dtype=nm.float64)
        return {'val': val}


functions = {
    'get_flux_value': (get_flux_value,),
}

fields = {
    'temperature': ('real', 1, 'Omega', 1),
}

variables = {
    'T': ('unknown field', 'temperature', 1, 1),  # 1 means history=1
    's': ('test field', 'temperature', 'T'),
}

ebcs = {
}

integrals = {
    'i': 2
}

equations = {
    'Temperature': """
      dw_dot.i.Omega_Cylinder(cylinder.rho_cp, s, dT/dt )
    + dw_dot.i.Omega_Plate(plate.rho_cp, s, dT/dt )
    + dw_dot.i.Omega_Powder(powder.rho_cp, s, dT/dt )
    + dw_laplace.i.Omega_Cylinder(cylinder.lam, s, T)
    + dw_laplace.i.Omega_Plate(plate.lam, s, T)
    + dw_laplace.i.Omega_Powder(powder.lam, s, T)
    = dw_integrate.i.Gamma_Source(heat_flux_defined_by_func.val, s)
    + dw_bc_newton.i.Gamma_Source(heat_loss.h_top, heat_loss.T_top_inf, s, T)
    + dw_bc_newton.i.Gamma_Plate(heat_loss.h_bot, heat_loss.T_bot_inf, s, T)
    """
}

ics = {
    'ic1': ('Omega_Powder', {'T.0': T0}),
    'ic2': ('Omega_Plate', {'T.0': T0}),
    'ic3': ('Omega_Cylinder', {'T.0': T0}),
}


def gen_probe():
    """Instantiates a line probe used later by the `step_hook` function."""
    p0, p1 = nm.array([0., 0., -10. * mm]), nm.array([0.0, 0.0, 15. * mm])
    line_probe = LineProbe(p0, p1, n_point=100, share_geometry=True)
    return line_probe


line_probe = gen_probe()
# inits an empty list that will hold the probe results
probe_results = []


def step_hook(pb, ts, variables):
    """
    This implements a function that gets called at every step from the
    time-solver.
    """
    T_field = pb.get_variables()['T']
    pars, vals = line_probe(T_field)
    probe_results.append(vals)


def post_process_hook(out, pb, state, extend=False):
    ts = pb.ts
    if ts.step == ts.n_step - 1:
        fig, (ax1, ax2) = plt.subplots(nrows=2)
        temperature_image = nm.array(probe_results).squeeze()
        m = ax1.imshow(temperature_image.T, origin='lower', aspect='auto')
        ax1.set_xlabel("time step")
        ax1.set_ylabel("distance across build\nplate and cylinder")
        fig.colorbar(m, ax=ax1, label="temperature")
        ax2.plot(temperature_image.T[0], label="bottom")
        ax2.plot(temperature_image.T[-1], label="top")
        ax2.set_xlabel("time step")
        ax2.set_ylabel("temperature (°C)")
        ax2.legend()
        fig.tight_layout()
        fig.savefig(os.path.join(pb.output_dir, 'heat_probe_time_evolution.png'),
                    bbox_inches='tight')
    return out


solvers = {
    'ls': ('ls.auto_direct', {
        # Reuse the factorized linear system from the first time step.
        'use_presolve': True,
        # Speed up the above by omitting the matrix digest check used normally
        # for verification that the current matrix corresponds to the
        # factorized matrix stored in the solver instance. Use with care!
        'use_mtx_digest': False,
    }),
    'newton': ('nls.newton', {
        'i_max': 1,
        'eps_a': 1e-5,
        'is_linear': True,
    }),
    'ts': ('ts.simple', {
        't0': 0.0,
        't1': 60.,
        'dt': None,
        'n_step': 120,
        'verbose': True,
        'is_linear': True,
    }),
}

options = {
    'step_hook': 'step_hook',
    'post_process_hook': 'post_process_hook',
}



import pytest

import sfepy.base.testing as tst

input_name = 'examples/homogenization/perfusion_micro.py'

def compare_scalars(s1, s2, l1= 's1', l2 = 's2',
                    allowed_error = 1e-8):

    diff  = abs(s1 - s2)
    tst.report( '|%s - %s|: %e' % (l1, l2, diff))
    if diff > allowed_error:
        return False
    else:
        return True

@pytest.mark.slow
def test_solution(output_dir):
    import sfepy
    from sfepy.base.base import Struct
    from sfepy.base.conf import ProblemConf, get_standard_keywords
    from sfepy.homogenization.homogen_app import HomogenizationApp
    import os.path as op

    ok = True

    required, other = get_standard_keywords()
    required.remove('equations')
    full_name = op.join(sfepy.base_dir, input_name)
    test_conf = ProblemConf.from_file(full_name, required, other)

    options = Struct(output_filename_trunk=None,
                     save_ebc=False,
                     save_ebc_nodes=False,
                     save_regions=False,
                     save_regions_as_groups=False,
                     solve_not=False)

    test_conf.options['output_dir'] = output_dir
    app = HomogenizationApp(test_conf, options, 'homogen:' )
    coefs = app()

    aerr = 1.0e-9
    tst.report('allowed error: abs = %e' % aerr)

    # G^A = G^B ?
    ok = ok and compare_scalars(coefs.GA, coefs.GB,\
                                'G^A', 'G^B', aerr)

    # F^{A+} + F^{B+} = -1/h \int_{\partial_+Y_m} ?
    aux = 1.0 / test_conf.param_h * coefs.Volume_bYMp
    ok = ok and compare_scalars(coefs.FpA + coefs.FpB, -aux,
                                'F^{A+} + F^{B+}', '-bYM^+', aerr)

    # F^{A-} + F^{B-} = -1/h \int_{\partial_-Y_m} ?
    aux = 1.0 / test_conf.param_h * coefs.Volume_bYMm
    ok = ok and compare_scalars(coefs.FmA + coefs.FmB, -aux,
                                'F^{A-} + F^{B-}', '-bYM^-', aerr)

    # symmetry of H ?
    ok = ok and compare_scalars(coefs.Hpm, coefs.Hmp,
                                'H^{+-}', 'H^{-+}', aerr)

    # E = -F ?
    ok = ok and compare_scalars(coefs.EmA, -coefs.FmA,
                                'E^{A-}', '-F^{A-}',aerr)
    ok = ok and compare_scalars(coefs.EpA, -coefs.FpA,
                                'E^{A+}', '-F^{A+}',aerr)
    ok = ok and compare_scalars(coefs.EmB, -coefs.FmB,
                                'E^{B-}', '-F^{B-}',aerr)
    ok = ok and compare_scalars(coefs.EpB, -coefs.FpB,
                                'E^{B+}', '-F^{B+}',aerr)

    # S = S_test ?
    coefsd = coefs.to_dict()
    compare = []
    for ii in coefsd.keys():
        if 'S_test' in ii:
            ch = ii[6]
            io = ii[-1]
            compare.append((ii, 'S%s_%s' % (ch, io)))

    for s1, s2 in compare:
        ok = ok and tst.compare_vectors(coefsd[s1], -coefsd[s2],
                                        label1='S_test', label2='S',
                                        allowed_error=aerr)

    assert ok



from sfepy.base.base import assert_
import sfepy.base.testing as tst

def _cmp(s1, s2):
    s1 = s1.split()
    s2 = s2.split()
    v1, t1 = float(s1[0]), s1[1:]
    v2, t2 = float(s2[0]), s2[1:]
    return (abs(v1 - v2) < (1e-15 * abs(v1))) and (t1 == t2)

def test_units():
    from sfepy.mechanics.units import Unit, Quantity, sm

    if sm is None:
        tst.report('cannot import sympy, skipping')
        return

    units = ['m', 's', 'kg', 'C']
    tst.report('units:', units)
    unit_set = [Unit(key) for key in units]

    q1 = Quantity('stress', unit_set)
    tst.report(q1.name, ':', q1())
    assert_(_cmp(q1(), '1.0 Pa'))
    assert_(_cmp(q1('c'), '100.0 cPa'))

    q2 = Quantity('force', unit_set)
    tst.report(q2.name, ':', q2())
    assert_(_cmp(q2(), '1.0 Newton'))
    assert_(_cmp(q2('d'), '0.1 dNewton'))

    q3 = Quantity('energy', unit_set)
    tst.report(q3.name, ':', q3())
    assert_(_cmp(q3(), '1.0 J'))
    assert_(_cmp(q3('mu'), '1000000.0 muJ'))

    units = ['mm', 's', 'g', 'C']
    tst.report('units:', units)
    unit_set = [Unit(key) for key in units]

    q1 = Quantity('stress', unit_set)
    tst.report(q1.name, ':', q1())
    assert_(_cmp(q1(), '1.0 Pa'))

    q2 = Quantity('force', unit_set)
    tst.report(q2.name, ':', q2())
    assert_(_cmp(q2(), '1.0 muNewton'))

    q3 = Quantity('energy', unit_set)
    tst.report(q3.name, ':', q3())
    assert_(_cmp(q3(), '1.0 nJ'))

    units = ['cm', 'ms', 'kg', 'kC']
    tst.report('units:', units)
    unit_set = [Unit(key) for key in units]

    q1 = Quantity('stress', unit_set)
    tst.report(q1.name, ':', q1())
    assert_(_cmp(q1(), '0.1 GPa'))

    q2 = Quantity('force', unit_set)
    tst.report(q2.name, ':', q2())
    assert_(_cmp(q2(), '10.0 kNewton'))

    q3 = Quantity('energy', unit_set)
    tst.report(q3.name, ':', q3())
    assert_(_cmp(q3(), '0.1 kJ'))

    q4 = Quantity('thermal_expandability', unit_set)
    tst.report(q4.name, ':', q4())
    assert_(_cmp(q4(), '0.1 MPa / C'))

    assert_(_cmp(q4('G'), '0.0001 GPa / C'))
    assert_(_cmp(q4('M'), '0.1 MPa / C'))
    assert_(_cmp(q4('k'), '100.0 kPa / C'))
    assert_(_cmp(q4('d'), '10000.0 dPa / C'))
    assert_(_cmp(q4(''), '100000.0 Pa / C'))

    units = ['m', 's', 'g', 'C']
    tst.report('units:', units)
    unit_set = [Unit(key) for key in units]

    q4 = Quantity('thermal_expandability', unit_set)
    tst.report(q4.name, ':', q4())
    assert_(_cmp(q4(), '1.0 mPa / C'))

    assert_(_cmp(q4('k'), str(0.000001) + ' kPa / C'))
    assert_(_cmp(q4('d'), '0.0001 dPa / C'))
    assert_(_cmp(q4(''), '0.001 Pa / C'))
    assert_(_cmp(q4('c'), '0.1 cPa / C'))
    assert_(_cmp(q4('m'), '1.0 mPa / C'))
    assert_(_cmp(q4('mu'), '1000.0 muPa / C'))
    assert_(_cmp(q4('n'), '1000000.0 nPa / C'))

def test_consistent_sets():
    from sfepy.mechanics.units import get_consistent_unit_set, sm

    if sm is None:
        tst.report('cannot import sympy, skipping')
        return True

    u_sets = {
        ('m', 's', 'kg', 'C') : {'force' : '1.0 Newton',
                                 'stress' : '1.0 Pa',
                                 'energy' : '1.0 J',
                                 'thermal_expandability' : '1.0 Pa / C'},
        ('mm', 's', 'kg', 'C') : {'force' : '1.0 mNewton',
                                  'stress' : '1.0 kPa',
                                  'energy' : '1.0 muJ',
                                 'thermal_expandability' : '1.0 kPa / C'},
        ('mm', 's', 'g', 'C') : {'force' : '1.0 muNewton',
                                 'stress' : '1.0 Pa',
                                 'energy' : '1.0 nJ',
                                 'thermal_expandability' : '1.0 Pa / C'},
    }

    ok = True
    for unit_set, true_derived_units in u_sets.items():
        tst.report('units:', unit_set)
        derived_units = get_consistent_unit_set(*unit_set)

        for key, true_val in true_derived_units.items():
            val = derived_units[key]
            _ok = _cmp(true_val, val)
            tst.report('%s: %s == %s -> %s' % (key, true_val, val, _ok))

            ok = ok and _ok

    assert_(ok)



from sfepy.discrete.fem import Mesh, FEDomain
import scipy.sparse as sps
import numpy as nm
from sfepy.base.compat import factorial
from sfepy.base.base import output
from sfepy.discrete.common.extmods.cmesh import (create_mesh_graph,
                                                 graph_components)

def elems_q2t(el):

    nel, nnd = el.shape
    if nnd > 4:
        q2t = nm.array([[0, 2, 3, 6],
                        [0, 3, 7, 6],
                        [0, 7, 4, 6],
                        [0, 5, 6, 4],
                        [1, 5, 6, 0],
                        [1, 6, 2, 0]])

    else:
        q2t = nm.array([[0, 1, 2],
                        [0, 2, 3]])

    ns, nn = q2t.shape
    nel *= ns

    out = nm.zeros((nel, nn), dtype=nm.int32);

    for ii in range(ns):
        idxs = nm.arange(ii, nel, ns)

        out[idxs,:] = el[:, q2t[ii,:]]

    return nm.ascontiguousarray(out)

def triangulate(mesh, verbose=False):
    """
    Triangulate a 2D or 3D tensor product mesh: quadrilaterals->triangles,
    hexahedrons->tetrahedrons.

    Parameters
    ----------
    mesh : Mesh
        The input mesh.

    Returns
    -------
    mesh : Mesh
        The triangulated mesh.
    """
    conns = None
    for k, new_desc in [('3_8', '3_4'), ('2_4', '2_3')]:
        if k in mesh.descs:
            conns = mesh.get_conn(k)
            break

    if conns is not None:
        nelo = conns.shape[0]
        output('initial mesh: %d elements' % nelo, verbose=verbose)

        new_conns = elems_q2t(conns)
        nn = new_conns.shape[0] // nelo
        new_cgroups = nm.repeat(mesh.cmesh.cell_groups, nn)

        output('new mesh: %d elements' % new_conns.shape[0], verbose=verbose)
        mesh = Mesh.from_data(mesh.name, mesh.coors,
                              mesh.cmesh.vertex_groups,
                              [new_conns], [new_cgroups], [new_desc])

    return mesh

def smooth_mesh(mesh, n_iter=4, lam=0.6307, mu=-0.6347,
                weights=None, bconstr=True,
                volume_corr=False):
    """
    FE mesh smoothing.

    Based on:

    [1] Steven K. Boyd, Ralph Muller, Smooth surface meshing for automated
    finite element model generation from 3D image data, Journal of
    Biomechanics, Volume 39, Issue 7, 2006, Pages 1287-1295,
    ISSN 0021-9290, 10.1016/j.jbiomech.2005.03.006.
    (http://www.sciencedirect.com/science/article/pii/S0021929005001442)

    Parameters
    ----------
    mesh : mesh
        FE mesh.
    n_iter : integer, optional
        Number of iteration steps.
    lam : float, optional
        Smoothing factor, see [1].
    mu : float, optional
        Unshrinking factor, see [1].
    weights : array, optional
        Edge weights, see [1].
    bconstr: logical, optional
        Boundary constraints, if True only surface smoothing performed.
    volume_corr: logical, optional
        Correct volume after smoothing process.

    Returns
    -------
    coors : array
        Coordinates of mesh nodes.
    """

    def laplacian(coors, weights):

        n_nod = coors.shape[0]
        displ = (weights - sps.identity(n_nod)) * coors

        return displ

    def taubin(coors0, weights, lam, mu, n_iter):

        coors = coors0.copy()

        for ii in range(n_iter):
            displ = laplacian(coors, weights)
            if nm.mod(ii, 2) == 0:
                coors += lam * displ
            else:
                coors += mu * displ

        return coors

    def get_volume(el, nd):
        from sfepy.linalg.utils import dets_fast

        dim = nd.shape[1]
        nnd = el.shape[1]

        etype = '%d_%d' % (dim, nnd)
        if etype == '2_4' or etype == '3_8':
            el = elems_q2t(el)

        nel = el.shape[0]

        #bc = nm.zeros((dim, ), dtype=nm.double)
        mul = 1.0 / factorial(dim)
        if dim == 3:
            mul *= -1.0

        mtx = nm.ones((nel, dim + 1, dim + 1), dtype=nm.double)
        mtx[:,:,:-1] = nd[el,:]
        vols = mul * dets_fast(mtx)
        vol = vols.sum()
        bc = nm.dot(vols, mtx.sum(1)[:,:-1] / nnd)

        bc /= vol

        return vol, bc

    from sfepy.base.timing import Timer

    output('smoothing...')
    timer = Timer(start=True)

    if weights is None:
        n_nod = mesh.n_nod
        domain = FEDomain('mesh', mesh)
        cmesh = domain.cmesh

        # initiate all vertices as inner - hierarchy = 2
        node_group = nm.ones((n_nod,), dtype=nm.int16) * 2
        # boundary vertices - set hierarchy = 4
        if bconstr:
            # get "vertices of surface"
            facets = cmesh.get_surface_facets()
            f_verts = cmesh.get_incident(0, facets, cmesh.dim - 1)
            node_group[f_verts] = 4

        # generate costs matrix
        e_verts = cmesh.get_conn(1, 0).indices
        fc1, fc2 = e_verts[0::2], e_verts[1::2]
        idxs = nm.where(node_group[fc2] >= node_group[fc1])
        rows1 = fc1[idxs]
        cols1 = fc2[idxs]
        idxs = nm.where(node_group[fc1] >= node_group[fc2])
        rows2 = fc2[idxs]
        cols2 = fc1[idxs]
        crows = nm.concatenate((rows1, rows2))
        ccols = nm.concatenate((cols1, cols2))
        costs = sps.coo_matrix((nm.ones_like(crows), (crows, ccols)),
                               shape=(n_nod, n_nod),
                               dtype=nm.double)

        # generate weights matrix
        idxs = list(range(n_nod))
        aux = sps.coo_matrix((1.0 / nm.asarray(costs.sum(1)).squeeze(),
                              (idxs, idxs)),
                             shape=(n_nod, n_nod),
                             dtype=nm.double)

        #aux.setdiag(1.0 / costs.sum(1))
        weights = (aux.tocsc() * costs.tocsc()).tocsr()

    coors = taubin(mesh.coors, weights, lam, mu, n_iter)

    output('...done in %.2f s' % timer.stop())

    if volume_corr:
        output('rescaling...')
        timer.start()
        volume0, bc = get_volume(mesh.conns[0], mesh.coors)
        volume, _ = get_volume(mesh.conns[0], coors)

        scale = volume0 / volume
        output('scale factor: %.2f' % scale)

        coors = (coors - bc) * scale + bc

        output('...done in %.2f s' % timer.stop())

    return coors

def expand2d(mesh2d, dist, rep):
    """
    Expand 2D planar mesh into 3D volume,
    convert triangular/quad mesh to tetrahedrons/hexahedrons.

    Parameters
    ----------
    mesh2d : Mesh
        The 2D mesh.
    dist : float
        The elements size in the 3rd direction.
    rep : int
        The number of elements in the 3rd direction.

    Returns
    -------
    mesh3d : Mesh
        The 3D mesh.
    """
    if len(mesh2d.descs) > 1:
        raise ValueError('More than one cell type (%s). Not supported!'
                         % ', '.join(mesh2d.descs))

    nel = mesh2d.n_el
    nnd = mesh2d.n_nod
    et = mesh2d.descs[0]
    coors = mesh2d.coors
    conn = mesh2d.get_conn(et)

    zcoor = nm.arange(rep + 1) * dist
    coors3d = nm.hstack([nm.tile(coors, (rep + 1, 1)),
                         nm.tile(zcoor, (nnd,1)).T.flatten()[:,nm.newaxis]])
    ngroups = nm.tile(mesh2d.cmesh.vertex_groups, (rep + 1,))

    if et == '2_4':
        descs3d = '3_8'
        conn3d = nm.zeros((nel * rep, 8), dtype=nm.int32)
        mats3d = nm.tile(mesh2d.cmesh.cell_groups, (1, rep)).squeeze()

    elif et == '2_3':
        descs3d = '3_4'
        conn3d = nm.zeros((3 * nel * rep, 4), dtype=nm.int32)
        mats3d = nm.tile(mesh2d.cmesh.cell_groups, (1, 3 * rep)).squeeze()

    for ii in range(rep):
        bgn0 = nnd * ii
        bgn1 = bgn0 + nnd
        if et == '2_4':
            bge0 = nel * ii
            bge1 = bge0 + nel
            conn3d[bge0:bge1,:4] = conn + bgn0
            conn3d[bge0:bge1,4:] = conn + bgn1

        elif et == '2_3':
            # 0 1 2 5
            bge0 = 3 * nel * ii
            bge1 = bge0 + nel
            conn3d[bge0:bge1,:] = nm.array([conn[:,0] + bgn0,
                                            conn[:,1] + bgn0,
                                            conn[:,2] + bgn0,
                                            conn[:,2] + bgn1]).T
            # 0 1 5 4
            bge0 += nel
            bge1 += nel
            conn3d[bge0:bge1,:] = nm.array([conn[:,0] + bgn0,
                                            conn[:,1] + bgn0,
                                            conn[:,2] + bgn1,
                                            conn[:,1] + bgn1]).T
            # 0 4 5 3
            bge0 += nel
            bge1 += nel
            conn3d[bge0:bge1,:] = nm.array([conn[:,0] + bgn0,
                                            conn[:,1] + bgn1,
                                            conn[:,2] + bgn1,
                                            conn[:,0] + bgn1]).T

    mesh3d = Mesh.from_data('mesh', coors3d, ngroups, [conn3d],
                            [mats3d], [descs3d])

    return mesh3d

def merge_lines(mesh, eps=1e-18):
    """
    Merge edges of an edge-only mesh that are in the same direction w.r.t. the
    tolerance `eps`.
    """
    coors, ngroups, conns, mat_ids, ctype = mesh
    conns = conns[0]

    # vertices to edges map
    n_v = coors.shape[0]
    n_e = conns.shape[0]
    row = nm.repeat(nm.arange(n_e), 2)
    aux = sps.coo_matrix((nm.ones((n_e * 2,), dtype=bool),
                          (row, conns.flatten())), shape=(n_e, n_v))
    v2e = aux.tocsc()
    n_epv = nm.diff(v2e.indptr)

    # directional vectors of edges
    de = coors[conns[:, 1], :] - coors[conns[:, 0], :]
    de = de / nm.linalg.norm(de, axis=1)[:, nm.newaxis]

    eflag = nm.ones((n_e, ), dtype=bool)
    valid_e = nm.where(eflag)[0]
    e_remove = []

    while len(valid_e) > 0:
        ie = valid_e[0]
        d = de[ie]
        buff = [(ie, conns[ie, 0]), (ie, conns[ie, 1])]
        eflag[ie] = False  # invalidate edge
        while len(buff) > 0:
            e, v = buff.pop(-1)
            if n_epv[v] == 2:
                idx = v2e.indptr[v]
                aux = v2e.indices[idx]
                next_e = v2e.indices[idx + 1] if aux == e else aux

                if not eflag[next_e]:  # valid edge?
                    continue
                if nm.linalg.norm(de[next_e] - d) < eps\
                        or nm.linalg.norm(de[next_e] + d) < eps:
                    next_ec = conns[next_e, :]
                    new_v = next_ec[0] if next_ec[1] == v else next_ec[1]
                    idx = 0 if conns[e, 0] == v else 1
                    conns[e, idx] = new_v  # reconnect edge

                    idx = v2e.indptr[new_v]
                    aux = v2e.indices[idx]
                    idx += 0 if aux == next_e else 1
                    v2e.indices[idx] = e  # update v2e map

                    buff.append((e, new_v))  # continue in searching
                    eflag[next_e] = False  # invalidate edge
                    e_remove.append(next_e)

        valid_e = nm.where(eflag)[0]

    if len(e_remove) > 0:
        # remove unused edges and vertices
        eflag.fill(True)
        eflag[nm.asarray(e_remove)] = False
        remap = -nm.ones((n_v, ), dtype=nm.int64)
        remap[conns[eflag, :]] = 1
        vidx = nm.where(remap > 0)[0]
        remap[vidx] = nm.arange(len(vidx))
        conns_new = remap[conns[eflag, :]]

        return coors[vidx, :], ngroups[vidx],\
            [conns_new], [mat_ids[0][eflag]], ctype
    else:
        return mesh

def extract_edges(mesh, eps=1e-16):
    """
    Extract outline edges of a given mesh.
    The outline edge is an edge for which norm(nvec_1 - nvec_2) < eps,
    where nvec_1 and nvec_2 are the normal vectors of the incident facets.

    Parameters
    ----------
    mesh : Mesh
        The 3D or 2D mesh.
    eps : float
        The tolerance parameter of the outline edge searching algorithm.

    Returns
    -------
    mesh_out : tuple
        The data of the outline mesh, Mesh.from_data() format, i.e.
        (coors, ngroups, ed_conns, mat_ids, descs).
    """
    domain = FEDomain('domain', mesh)
    cmesh = domain.cmesh

    output('Mesh - dimension: %d, vertices: %d, elements: %d'
           % (mesh.dim, mesh.n_nod, mesh.n_el))

    if mesh.dim == 2:
        oedges = cmesh.get_surface_facets()
        mesh_coors = nm.hstack([cmesh.coors,
                                nm.zeros((cmesh.coors.shape[0], 1))])

    elif mesh.dim == 3:
        cmesh.setup_connectivity(1, 2)
        cmesh.setup_connectivity(3, 2)

        sfaces = cmesh.get_surface_facets()
        _, idxs = nm.unique(cmesh.get_conn(3, 2).indices, return_index=True)

        normals = cmesh.get_facet_normals()[idxs, :]
        se_map, se_off = cmesh.get_incident(1, sfaces, 2, ret_offsets=True)
        sedges = nm.unique(se_map)
        n_se = sedges.shape[0]

        # remap surface edges to continuous range
        se_remap = -nm.ones(sedges.max() + 1)
        se_remap[sedges] = nm.arange(n_se)
        se_map0 = se_remap[se_map]

        # surface face/edge connectivity matrix (n_surf x n_edge)
        n_ef = nm.diff(se_off)[0]  # = 2
        n_sf = se_map.shape[0] // n_ef
        row = nm.repeat(nm.arange(n_sf), n_ef)
        sf2e = sps.coo_matrix((nm.ones((n_sf * n_ef,), dtype=bool),
                               (row , se_map0)), shape=(n_sf, n_se))
        # edge to face map (n_edge x 2)
        se2f = sf2e.tocsc().indices.reshape((sedges.shape[0], 2))

        snormals = normals[sfaces]
        err = nm.linalg.norm(snormals[se2f[:, 0]] - snormals[se2f[:, 1]],
                             axis=1)
        oedges = sedges[nm.where(err > eps)[0]]
        mesh_coors = cmesh.coors

    else:
        raise NotImplementedError

    # save outline mesh
    if oedges.shape[0] > 0:
        ec_idxs = nm.unique(cmesh.get_incident(0, oedges, 1))
        ed_coors = mesh_coors[ec_idxs, :]
        ngroups = nm.zeros((ed_coors.shape[0],), dtype=nm.int16)

        aux = cmesh.get_conn(1, 0).indices

        ed_conns = aux.reshape((aux.shape[0] // 2, 2))[oedges, :]
        ec_remap = -nm.ones((ec_idxs.max() + 1, ), dtype=nm.int64)
        ec_remap[ec_idxs] = nm.arange(ec_idxs.shape[0])
        ed_conns = ec_remap[ed_conns]
        mat_ids = nm.ones((ed_conns.shape[0],), dtype=nm.int16)

        mesh_out = ed_coors, ngroups, [ed_conns], [mat_ids], ['3_2']

        return mesh_out

    else:
        raise ValueError('no outline edges found (eps=%e)!' % eps)

def _get_facets(vertices, offsets, ii, n_fp):
    facets = []
    for ic in range(n_fp):
        facets.append(vertices[offsets[ii] + ic][:, None])

    facets = nm.concatenate(facets, axis=1)

    return nm.ascontiguousarray(facets.astype(nm.int32))

def get_surface_faces(domain):
    cmesh = domain.cmesh
    faces = cmesh.get_surface_facets()
    vertices_f, offs_f = cmesh.get_incident(0, faces,
                                            cmesh.dim - 1, ret_offsets=True)

    n_fp = nm.diff(offs_f)
    surf_faces = []

    itri = nm.where(n_fp == 3)[0]
    if itri.size:
        surf_faces.append(_get_facets(vertices_f, offs_f, itri, 3))

    itet = nm.where(n_fp == 4)[0]
    if itet.size:
        surf_faces.append(_get_facets(vertices_f, offs_f, itet, 4))

    cells_c, offs_c = cmesh.get_incident(cmesh.dim, faces, cmesh.dim - 1,
                                         ret_offsets=True)
    ids = cmesh.get_local_ids(faces, cmesh.dim - 1, cells_c, offs_c,
                              cmesh.dim)
    lst = nm.c_[cells_c, ids]

    return lst, surf_faces

def surface_graph(surf_faces, n_nod):
    nnz, prow, icol = create_mesh_graph(n_nod, n_nod, len(surf_faces),
                                        surf_faces, surf_faces)
    data = nm.empty((nnz,), dtype=nm.int32)
    data.fill(2)
    return sps.csr_matrix((data, icol, prow), (n_nod, n_nod))

def surface_components(gr_s, surf_faces):
    """
    Determine surface components given surface mesh connectivity graph.
    """
    n_nod = gr_s.shape[0]
    n_comp, flag = graph_components(n_nod, gr_s.indptr, gr_s.indices)

    comps = []
    for ii, face in enumerate(surf_faces):
        comp = flag[face[:,0]]
        comps.append(comp)

    return n_comp, comps



import os
import numpy as nm

import sfepy.base.testing as tst

def test_linearization(output_dir):
    from sfepy.base.base import Struct
    from sfepy.discrete.fem import Mesh, FEDomain, Field
    from sfepy import data_dir

    geometries = ['2_3', '2_4', '3_4', '3_8']
    approx_orders = [1, 2]
    funs = [nm.cos, nm.sin, lambda x: x]

    ok = True
    for geometry in geometries:
        name = os.path.join(data_dir,
                            'meshes/elements/%s_1.mesh' % geometry)
        mesh = Mesh.from_file(name)

        domain = FEDomain('', mesh)
        domain = domain.refine()

        domain.mesh.write(os.path.join(output_dir,
                                       'linearizer-%s-0.mesh' % geometry))

        omega = domain.create_region('Omega', 'all')

        for approx_order in approx_orders:
            for dpn in [1, mesh.dim]:
                tst.report('geometry: %s, approx. order: %d, dpn: %d' %
                           (geometry, approx_order, dpn))

                field = Field.from_args('fu', nm.float64, dpn, omega,
                                        approx_order=approx_order)

                cc = field.get_coor()
                dofs = nm.zeros((field.n_nod, dpn), dtype=nm.float64)

                for ic in range(dpn):
                    dofs[:, ic] = funs[ic](3 * (cc[:, 0] * cc[:, 1]))

                vmesh, vdofs, level = field.linearize(dofs,
                                                      min_level=0,
                                                      max_level=3,
                                                      eps=1e-2)

                if approx_order == 1:
                    _ok = level == 0

                else:
                    _ok = level > 0
                tst.report('max. refinement level: %d: %s' % (level, _ok))

                ok = ok and _ok

                rdofs = nm.zeros((vmesh.n_nod, dpn), dtype=nm.float64)
                cc = vmesh.coors
                for ic in range(dpn):
                    rdofs[:, ic] = funs[ic](3 * (cc[:, 0] * cc[:, 1]))

                _ok = nm.allclose(rdofs, vdofs, rtol=0.0, atol=0.03)
                tst.report('interpolation: %s' % _ok)
                ok = ok and _ok

                out = {
                    'u' : Struct(name='output_data',
                                 mode='vertex', data=vdofs,
                                 var_name='u', dofs=None)
                }

                name = os.path.join(output_dir,
                                    'linearizer-%s-%d-%d'
                                    % (geometry, approx_order, dpn))

                vmesh.write(name + '.mesh')
                vmesh.write(name + '.vtk', out=out)

    assert ok



import numpy as nm
import pytest

try:
    import sfepy.linalg.sympy_operators as sops
except ImportError:
    sops = None

from sfepy import data_dir
import sfepy.base.testing as tst

filename_mesh = data_dir + '/meshes/2d/special/circle_in_square.mesh'

dim = 2

field_1 = {
    'name' : 'a_harmonic_field',
    'dtype' : 'real',
    'shape' : 'scalar',
    'region' : 'Omega',
    'approx_order' : 1,
}

variables = {
    't': ('unknown field', 'a_harmonic_field', 0),
    's': ('test field',    'a_harmonic_field', 't'),
}

regions = {
    'Omega' : 'all',
    'Gamma' : ('vertices of surface', 'facet'),
}

ebcs = {
    't_left' : ('Gamma', {'t.0' : 'ebc'}),
}

integral_1 = {
    'name' : 'i',
    'order' : 2,
}

material_1 = {
    'name' : 'coef',
    'values' : {
        'val' : 12.0,
        'K' : [[1.0, 0.3], [0.3, 2.0]],
    }
}

material_2 = {
    'name' : 'rhs',
    'function' : 'rhs',
}

equations = {
    'Laplace' :
    """2 * dw_laplace.i.Omega(coef.val, s, t)
    """,
    'Diffusion' :
    """3 * dw_diffusion.i.Omega(coef.K, s, t)
    """,
}
equations_rhs = {
    'Laplace' :
    """= - dw_volume_lvf.i.Omega(rhs.val, s)""",
    'Diffusion' :
    """= - dw_volume_lvf.i.Omega(rhs.val, s)""",
}

solutions = {
    'sincos' : ('t', 'sin(3.0 * x) * cos(4.0 * y)'),
    'poly' : ('t', '(x**2) + (y**2)'),
    'polysin' : ('t', '((x - 0.5)**3) * sin(5.0 * y)'),
}

solver_0 = {
    'name' : 'ls',
    'kind' : 'ls.scipy_direct',
}

solver_1 = {
    'name' : 'newton',
    'kind' : 'nls.newton',

    'i_max'      : 1,
    'eps_a'      : 1e-10,
}

solution = ['']
def ebc(ts, coor, solution=None):
    expression = solution[0]
    val = tst.eval_coor_expression(expression, coor)
    return nm.atleast_1d(val)

def rhs(ts, coor, mode=None, expression=None, **kwargs):
    if mode == 'qp':
        if expression is None:
            expression = '0.0 * x'

        val = tst.eval_coor_expression(expression, coor)
        val.shape = (val.shape[0], 1, 1)
        return {'val' : val}

functions = {
    'ebc' : (lambda ts, coor, **kwargs:
             ebc(ts, coor, solution=solution),),
    'rhs' : (rhs,),
}

def _build_rhs(equation, sols):
    rhss = {}
    tst.report('%s:' % equation.name)
    tst.report('evaluating terms, "<=" is solution, "=>" is the rhs:')
    for term in equation.terms:
        if not hasattr(term, 'symbolic'):
            tst.report('term %s has no symbolic description!' % term.name)
            raise ValueError
        expr = term.symbolic['expression']
        arg_map = term.symbolic['map']
        tst.report('%s(%s)' % (term.name, ', '.join(term.ats)))
        tst.report('multiplicator: %f' % term.sign)
        tst.report('  symbolic:', expr)
        tst.report('  using argument map:', arg_map)

        for sol_name, sol in sols.items():
            rhs = _eval_term(sol[1], term, sops)
            srhs = "(%s * (%s))" % (term.sign, rhs)
            rhss.setdefault(sol_name, []).append(srhs)

    for key, val in rhss.items():
        rhss[key] = '+'.join(val)

    return rhss

def _eval_term(sol, term, sops):
    """Works for scalar, single unknown terms only!"""
    expr = term.symbolic['expression']
    arg_map = term.symbolic['map']
    env = {'x' : sops.Symbol('x'),
           'y' : sops.Symbol('y'),
           'z' : sops.Symbol('z'),
           'dim' : dim}
    for key, val in arg_map.items():
        if val == 'state':
            env[key] = sol
        else:
            env[key] = term.get_args([val])[0]

        if 'material' in val:
            # Take the first value - constant in all QPs.
            aux = env[key][0,0]
            if nm.prod(aux.shape) == 1:
                env[key] = aux.squeeze()
            else:
                import sympy
                env[key] = sympy.Matrix(aux)

    tst.report('  <= ', sol)
    sops.set_dim(dim)
    val = str(eval(expr, sops.__dict__, env))
    tst.report('   =>', val)
    return val

def _test_msm_symbolic(problem, equations, output_dir):
    import os.path as op

    if sops is None:
        tst.report('cannot import sympy, skipping')
        return True

    ok = True
    for eq_name, equation in equations.items():
        problem.set_equations({eq_name : equation})
        problem.update_materials()

        rhss = _build_rhs(problem.equations[eq_name], solutions)
        erhs = problem.conf.equations_rhs[eq_name]

        problem.set_equations({eq_name : equation + erhs})
        variables = problem.get_variables()
        materials = problem.get_materials()
        rhs_mat = materials['rhs']

        for sol_name, sol in problem.conf.solutions.items():
            tst.report('testing', sol_name)
            var_name, sol_expr = sol
            rhs_expr = rhss[sol_name]

            tst.report('sol:', sol_expr)
            tst.report('rhs:', rhs_expr)
            globals()['solution'][0] = sol_expr
            rhs_mat.function.set_extra_args(expression=rhs_expr)
            problem.time_update()
            state = problem.solve(save_results=False)
            coor = variables[var_name].field.get_coor()
            ana_sol = tst.eval_coor_expression(sol_expr, coor)
            num_sol = state(var_name)

            ana_norm = nm.linalg.norm(ana_sol, nm.inf)
            ret = tst.compare_vectors(ana_sol, num_sol,
                                      allowed_error=ana_norm * 1e-2,
                                      label1='analytical %s' % var_name,
                                      label2='numerical %s' % var_name,
                                      norm=nm.inf)
            if not ret:
                tst.report('variable %s: failed' % var_name)

            fname = op.join(output_dir, 'test_msm_symbolic_%s.vtk')
            out = {}
            astate = state.copy()
            astate.set_state(ana_sol)
            aux = astate.create_output()
            out['ana_t'] = aux['t']
            aux = state.create_output()
            out['num_t'] = aux['t']

            problem.domain.mesh.write(fname % '_'.join((sol_name, eq_name)),
                                      io='auto', out=out)

            ok = ok and ret

    return ok

@pytest.fixture(scope='module')
def problem():
    import sys
    from sfepy.discrete import Problem
    from sfepy.base.conf import ProblemConf

    conf = ProblemConf.from_dict(globals(), sys.modules[__name__])

    problem = Problem.from_conf(conf)
    return problem

def _get_equations(problem, name):
    """
    Choose a sub-problem from all equations.
    """
    return {name : problem.conf.equations[name]}

def test_msm_symbolic_laplace(problem, output_dir):
    equations = _get_equations(problem, 'Laplace')
    assert _test_msm_symbolic(problem, equations, output_dir)

def test_msm_symbolic_diffusion(problem, output_dir):
    equations = _get_equations(problem, 'Diffusion')
    assert _test_msm_symbolic(problem, equations, output_dir)



from sfepy.linalg import dot_sequences
from sfepy.terms.terms import Term, terms

class LinearVolumeForceTerm(Term):
    r"""
    Vector or scalar linear volume forces (weak form) --- a right-hand side
    source term.

    :Definition:

    .. math::
        \int_{\Omega} \ul{f} \cdot \ul{v} \mbox{ or } \int_{\Omega} f q

    :Arguments:
        - material : :math:`\ul{f}` or :math:`f`
        - virtual  : :math:`\ul{v}` or :math:`q`
    """
    name = 'dw_volume_lvf'
    arg_types = ('material', 'virtual')
    arg_shapes = [{'material' : 'D, 1', 'virtual' : ('D', None)},
                  {'material' : '1, 1', 'virtual' : (1, None)}]

    function = staticmethod(terms.dw_volume_lvf)

    def get_fargs(self, mat, virtual,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vg, _ = self.get_mapping(virtual)

        return mat, vg


class NonlinearVolumeForceTerm(Term):
    r"""
    The volume force term with the force given by
    a user supplied function of the state variable.

    :Definition:

    .. math::
        \int_{\Omega} q f(p)

    :Arguments:
        - fun : :math:`f(p)`
        - dfun : :math:`\partial f(p) / \partial p`
        - virtual : :math:`q`
        - state : :math:`p`

    """
    name = 'dw_volume_nvf'
    arg_types = ('fun', 'dfun', 'virtual', 'state')
    arg_shapes = {'fun'     : lambda x: x,
                  'dfun'    : lambda x: x,
                  'virtual' : (1, 'state'),
                  'state'   : 1}

    @staticmethod
    def function(out, out_qp, geo):
        status = geo.integrate(out, out_qp)
        return status

    def get_fargs(self, fun, dfun, var1, var2,
                  mode=None, term_mode=None, diff_var=None, **kwargs):
        vg1, _ = self.get_mapping(var1)
        vg2, _ = self.get_mapping(var2)

        if diff_var is None:
            geo = vg1
            val_qp = fun(self.get(var2, 'val'))
            out_qp = dot_sequences(vg1.bf, val_qp,'ATB')

        else:
            geo = vg1
            val_d_qp = dfun(self.get(var2, 'val'))
            out_qp = dot_sequences(vg1.bf, val_d_qp*vg2.bf,'ATB')

        return out_qp, geo



from __future__ import absolute_import

import numpy as nm
import scipy.sparse as sps

from sfepy.base.base import get_default, try_imports, Struct
from sfepy.base.timing import Timer
from sfepy.solvers.solvers import Solver, EigenvalueSolver
import six
from six.moves import range

def eig(mtx_a, mtx_b=None, n_eigs=None, eigenvectors=True,
        return_time=None, solver_kind='eig.scipy', **ckwargs):
    """
    Utility function that constructs an eigenvalue solver given by
    `solver_kind`, calls it and returns its output.
    """
    kwargs = {'name' : 'aux', 'kind' : solver_kind}
    kwargs.update(ckwargs)
    conf = Struct(**kwargs)
    solver = Solver.any_from_conf(conf)

    status = {}
    out = solver(mtx_a, mtx_b, n_eigs, eigenvectors, status)
    if return_time is not None:
        return_time[0] = status['time']

    return out

def standard_call(call):
    """
    Decorator handling argument preparation and timing for eigensolvers.
    """
    def _standard_call(self, mtx_a, mtx_b=None, n_eigs=None,
                       eigenvectors=None, status=None, conf=None, **kwargs):
        timer = Timer(start=True)

        conf = get_default(conf, self.conf)
        mtx_a = get_default(mtx_a, self.mtx_a)
        mtx_b = get_default(mtx_b, self.mtx_b)
        n_eigs = get_default(n_eigs, self.n_eigs)
        eigenvectors = get_default(eigenvectors, self.eigenvectors)
        status = get_default(status, self.status)

        if n_eigs == 0:
            result =  self._ret_zero(mtx_a, eigenvectors=eigenvectors)

        else:
            result = call(self, mtx_a, mtx_b, n_eigs, eigenvectors, status,
                          conf, **kwargs)

        elapsed = timer.stop()
        if status is not None:
            status['time'] = elapsed

        return result

    return _standard_call

class ScipyEigenvalueSolver(EigenvalueSolver):
    """
    SciPy-based solver for both dense and sparse problems.

    The problem is consirered sparse if `n_eigs` argument is not None.
    """
    name = 'eig.scipy'

    _parameters = [
        ('method', "{'eig', 'eigh', 'eigs', 'eigsh'}", 'eigs', False,
         """The method for solving general or symmetric eigenvalue problems:
            for dense problems :func:`eig()` or :func:`eigh()` can be used, for
            sparse problems :func:`eigs()` or :func:`eigsh()` should be
            used."""),
        ('which', "'LM' | 'SM' | 'LR' | 'SR' | 'LI' | 'SI'", 'SM', False,
         """Which eigenvectors and eigenvalues to find,
            see :func:`scipy.sparse.linalg.eigs()`
            or :func:`scipy.sparse.linalg.eigsh()`. For dense problmes,
            only 'LM' and 'SM' can be used"""),
        ('*', '*', None, False,
         'Additional parameters supported by the method.'),
    ]

    def __init__(self, conf, **kwargs):
        EigenvalueSolver.__init__(self, conf, **kwargs)

        import scipy.linalg as sla
        self.sla = sla

        aux = try_imports(['import scipy.sparse.linalg as ssla'],
                          'cannot import scipy sparse eigenvalue solvers!')
        self.ssla = aux['ssla']

    @standard_call
    def __call__(self, mtx_a, mtx_b=None, n_eigs=None, eigenvectors=None,
                 status=None, conf=None):
        kwargs = self.build_solver_kwargs(conf)

        if conf.method in ('eig', 'eigh'):
            mtx_a, mtx_b = self._to_array(mtx_a, mtx_b)
            if conf.method == 'eig':
                out = self.sla.eig(mtx_a, mtx_b, right=eigenvectors, **kwargs)

            else:
                out = self.sla.eigh(mtx_a, mtx_b,
                                    eigvals_only=not eigenvectors, **kwargs)

        else:
            eig = self.ssla.eigs if conf.method == 'eigs' else self.ssla.eigsh
            out = eig(mtx_a, M=mtx_b, k=n_eigs, which=conf.which,
                      return_eigenvectors=eigenvectors, **kwargs)

        if eigenvectors:
            eigs = out[0]

        else:
            eigs = out

        if nm.iscomplexobj(eigs):
            ii = nm.argsort(nm.linalg.norm(eigs[:, None], axis=1))

        else:
            ii = nm.argsort(eigs)

        if n_eigs is not None and (conf.method in ('eig', 'eigh')):
            if conf.which == 'SM':
                ii = ii[:n_eigs]

            elif conf.which == 'LM':
                ii = ii[:-n_eigs-1:-1]

            else:
                raise ValueError("only 'LM' or 'SM' can be used with dense"
                                 " problems! (%s)" % conf.which)

        if eigenvectors:
            mtx_ev = out[1][:, ii]
            out = (eigs[ii], mtx_ev)

        else:
            out = eigs[ii]

        return out

class ScipySGEigenvalueSolver(EigenvalueSolver):
    """
    SciPy-based solver for dense symmetric problems.
    """
    name = 'eig.sgscipy'

    def __init__(self, conf, **kwargs):
        EigenvalueSolver.__init__(self, conf, **kwargs)

        try:
            import scipy.linalg.lapack as llapack
        except ImportError:
            import scipy.lib.lapack as llapack

        self.llapack = llapack

    @standard_call
    def __call__(self, mtx_a, mtx_b=None, n_eigs=None, eigenvectors=None,
                 status=None, conf=None):
        """
        Notes
        -----
        Eigenvectors argument is ignored, as they are computed always.
        """
        ll = self.llapack

        mtx_a, mtx_b = self._to_array(mtx_a, mtx_b)
        if nm.iscomplexobj(mtx_a):
            if mtx_b is None:
                fun = ll.get_lapack_funcs(['heev'], arrays=(mtx_a,))[0]
            else:
                fun = ll.get_lapack_funcs(['hegv'], arrays=(mtx_a,))[0]
        else:
            if mtx_b is None:
                fun = ll.get_lapack_funcs(['syev'], arrays=(mtx_a,))[0]
            else:
                fun = ll.get_lapack_funcs(['sygv'], arrays=(mtx_a,))[0]

        if mtx_b is None:
            out = fun(mtx_a)
        else:
            out = fun(mtx_a, mtx_b)

        # Fix output order of scipy.linalg.lapack functions.
        if out[0].ndim == 2:
            out = (out[1], out[0]) + out[2:]

        if not eigenvectors:
            if n_eigs is None:
                out = out[0]
            else:
                out = out[0][:n_eigs]
        else:
            if n_eigs is None:
                out = out[:-1]
            else:
                out = (out[0][:n_eigs], out[1][:, :n_eigs])

        return out

class LOBPCGEigenvalueSolver(EigenvalueSolver):
    """
    SciPy-based LOBPCG solver for sparse symmetric problems.
    """
    name = 'eig.scipy_lobpcg'

    _parameters = [
        ('i_max', 'int', 20, False,
         'The maximum number of iterations.'),
        ('eps_a', 'float', None, False,
         'The absolute tolerance for the convergence.'),
        ('largest', 'bool', True, False,
         'If True, solve for the largest eigenvalues, otherwise the smallest.'),
        ('precond', '{dense matrix, sparse matrix, LinearOperator}',
         None, False,
         'The preconditioner.'),
    ]

    def __init__(self, conf, **kwargs):
        EigenvalueSolver.__init__(self, conf, **kwargs)

        from scipy.sparse.linalg import lobpcg
        self.lobpcg = lobpcg

    @standard_call
    def __call__(self, mtx_a, mtx_b=None, n_eigs=None, eigenvectors=None,
                 status=None, conf=None):

        if n_eigs is None:
            n_eigs = mtx_a.shape[0]
        else:
            n_eigs = min(n_eigs, mtx_a.shape[0])

        rng = nm.random.default_rng(12345)
        x = rng.normal(size=(mtx_a.shape[0], n_eigs))

        out = self.lobpcg(mtx_a, x, mtx_b,
                          M=conf.precond,
                          tol=conf.eps_a, maxiter=conf.i_max,
                          largest=conf.largest,
                          verbosityLevel=conf.verbose)

        if not eigenvectors:
            out = out[0]

        return out

def init_slepc_args():
    try:
        import sys, slepc4py

    except ImportError:
        return

    argv = [arg for arg in sys.argv if arg not in ['-h', '--help']]
    slepc4py.init(argv)

class SLEPcEigenvalueSolver(EigenvalueSolver):
    """
    General SLEPc eigenvalue problem solver.
    """
    name = 'eig.slepc'

    _parameters = [
        ('method', 'str', 'krylovschur', False,
         'The actual solver to use.'),
        ('problem', 'str', 'gnhep', False,
         """The problem type: Hermitian (hep), non-Hermitian (nhep), generalized
         Hermitian (ghep), generalized non-Hermitian (gnhep), generalized
         non-Hermitian with positive semi-definite B (pgnhep), and generalized
         Hermitian-indefinite (ghiep)."""),
        ('i_max', 'int', 20, False,
         'The maximum number of iterations.'),
        ('eps', 'float', None, False,
         'The convergence tolerance.'),
        ('conv_test', '{"abs", "rel", "norm", "user"}, ', 'abs', False,
         'The type of convergence test.'),
        ('which', """{'largest_magnitude', 'smallest_magnitude',
        'largest_real', 'smallest_real',
        'largest_imaginary', 'smallest_imaginary', 'target_magnitude',
        'target_real', 'target_imaginary', 'all', 'which_user'}""",
         'largest_magnitude', False,
         'Which eigenvectors and eigenvalues to find.'),
        ('*', '*', None, False,
         'Additional parameters supported by the method.'),
    ]

    def __init__(self, conf, comm=None, context=None, **kwargs):
        if comm is None:
            init_slepc_args()

        from petsc4py import PETSc as petsc
        from slepc4py import SLEPc as slepc

        EigenvalueSolver.__init__(self, conf, petsc=petsc, slepc=slepc,
                                  comm=comm, context=context, **kwargs)

    def create_eps(self, options=None, comm=None):
        optDB = self.petsc.Options()

        if options is not None:
            for key, val in six.iteritems(options):
                optDB[key] = val

        es = self.slepc.EPS()
        es.create(comm)

        return es

    def create_petsc_matrix(self, mtx, comm=None):
        if mtx is None or isinstance(mtx, self.petsc.Mat):
            pmtx = mtx

        else:
            mtx = sps.csr_matrix(mtx)

            pmtx = self.petsc.Mat()
            pmtx.createAIJ(mtx.shape, csr=(mtx.indptr, mtx.indices, mtx.data),
                           comm=comm)

        return pmtx

    @standard_call
    def __call__(self, mtx_a, mtx_b=None, n_eigs=None, eigenvectors=None,
                 status=None, conf=None, comm=None, context=None):
        solver_kwargs = self.build_solver_kwargs(conf)

        pmtx_a = self.create_petsc_matrix(mtx_a, comm=comm)
        pmtx_b = self.create_petsc_matrix(mtx_b, comm=comm)

        es = self.create_eps(options=solver_kwargs, comm=comm)
        es.setType(conf.method)
        es.setProblemType(getattr(es.ProblemType, conf.problem.upper()))
        es.setDimensions(nev=n_eigs)
        es.setTolerances(tol=conf.eps, max_it=conf.i_max)
        es.setOperators(pmtx_a, pmtx_b)
        es.setConvergenceTest(getattr(es.Conv, conf.conv_test.upper()))
        es.setWhichEigenpairs(getattr(es.Which, conf.which.upper()))
        es.setFromOptions()

        es.solve()

        n_converged = es.getConverged()
        if status is not None:
            status['n_iter'] = es.getIterationNumber()
            status['n_converged'] = n_converged

        if not eigenvectors:
            out = nm.array([es.getEigenvalue(ii) for ii in range(n_converged)])

        else:
            vr, vi = pmtx_a.createVecs()
            eigs = []
            vrs, vis = [], []
            is_real = True
            for ii in range(n_converged):
                val = es.getEigenpair(ii, vr, vi)
                eigs.append(val if val.imag != 0 else val.real)
                vrs.append(vr.getArray().copy())
                vis.append(vi.getArray().copy())
                if is_real and nm.sum(nm.abs(vis[-1])) > 0.0:
                    is_real = False

            eigs = nm.array(eigs)
            vecs = nm.array(vrs).T
            if not is_real:
                vecs += 1j * nm.array(vis)

            out = (eigs, vecs)

        return out

class MatlabEigenvalueSolver(EigenvalueSolver):
    """
    Matlab eigenvalue problem solver.
    """
    name = 'eig.matlab'

    _parameters = [
        ('method', """{'eig', 'eigs', None}""", 'eigs', False,
         """The solution method. Note that eig() function cannot be used for
         all inputs. If `n_eigs` is not None, eigs() is used regardless of
         this parameter."""),
        ('balance', """{'balance', 'nobalance'}""", 'balance', False,
         'The balance option for eig().'),
        ('algorithm', """{'chol', 'qz'}""", 'chol', False,
         'The algorithm option for eig().'),
        ('which',
         """{'lm', 'sm', 'la', 'sa', 'be' 'lr', 'sr', 'li', 'si', sigma}""",
         'lm', False,
         'Which eigenvectors and eigenvalues to find with eigs().'),
        ('*', '*', None, False,
         'Additional parameters supported by eigs().'),
    ]

    def __init__(self, conf, comm=None, context=None, **kwargs):
        import matlab.engine as me

        EigenvalueSolver.__init__(self, conf, me=me, context=context,
                                  **kwargs)

    @standard_call
    def __call__(self, mtx_a, mtx_b=None, n_eigs=None, eigenvectors=None,
                 status=None, conf=None, comm=None, context=None):
        import os
        import shutil
        import tempfile
        import scipy.io as sio

        solver_kwargs = self.build_solver_kwargs(conf)

        dirname = tempfile.mkdtemp()
        mtx_filename = os.path.join(dirname, 'matrices.mat')
        eigs_filename = os.path.join(dirname, 'eigs.mat')
        sio.savemat(mtx_filename, {
            'A' : mtx_a,
            'B' : mtx_b if mtx_b is not None else 'None',
            'n_eigs' : n_eigs if n_eigs is not None else 'None',
            'eigenvectors' : (eigenvectors
                              if eigenvectors is not None else False),
            'method' : conf.method,
            'balance' : conf.balance,
            'algorithm' : conf.algorithm,
            'which' : conf.which,
            'verbose' : conf.verbose,
            'eigs_options' : solver_kwargs,
        })

        eng = self.me.start_matlab()
        eng.cd(os.path.dirname(__file__))
        eng.matlab_eig(mtx_filename, eigs_filename)
        eng.quit()

        evp = sio.loadmat(eigs_filename)

        shutil.rmtree(dirname)

        out = evp['vals'][:, 0]
        if eigenvectors:
            out =  (out, evp['vecs'])

        return out


class PrimmeEigenvalueSolver(EigenvalueSolver):
    """
    PRIMME eigenvalue problem solver.

    https://github.com/primme/primme

    Installation: pip install primme
    """
    name = 'eig.primme'

    _parameters = [
        ('which',
         "{'LM', 'SM', 'LA', 'SA', 'CLT', 'CGT'}", 'LM', False,
         'Which eigenvectors and eigenvalues to find.'),
        ('sigma', 'float', None, False,
         'Find eigenvalues near sigma.'),
        ('maxiter', 'int', None, False, 'Maximum number of iterations.'),
        ('tol', 'float', 0, False,
         'Tolerance for eigenpairs (stopping criterion).'),
        ('*', '*', None, False,
         'Additional parameters supported by eigsh().'),
    ]

    def __init__(self, conf, comm=None, context=None, **kwargs):
        import primme

        EigenvalueSolver.__init__(self, conf, primme=primme, context=context,
                                  **kwargs)

    @standard_call
    def __call__(self, mtx_a, mtx_b=None, n_eigs=None, eigenvectors=None,
                 status=None, conf=None, comm=None, context=None):
        if n_eigs is None:
            n_eigs = mtx_a.shape[0]

        solver_kwargs = self.build_solver_kwargs(conf)

        out = self.primme.eigsh(mtx_a, n_eigs, M=mtx_b,
                                which=conf.which.upper(),
                                tol=conf.tol,
                                maxiter=conf.maxiter,
                                sigma=conf.sigma,
                                return_eigenvectors=eigenvectors,
                                **solver_kwargs)

        return out



from __future__ import print_function
from __future__ import absolute_import
import os

import numpy as nm

from sfepy.base.base import get_default, Struct, output
from sfepy.base.ioutils import get_print_info
from sfepy.base.timing import Timer
from sfepy.discrete.fem import extend_cell_data, Mesh
from sfepy.homogenization.utils import coor_to_sym
from sfepy.base.conf import get_standard_keywords
from sfepy.discrete import Problem, Region
from sfepy.base.conf import ProblemConf
from sfepy.homogenization.coefficients import Coefficients
from sfepy.homogenization.micmac import get_correctors_from_file_hdf5
import os.path as op
import six
from six.moves import range
import atexit

shared = Struct()

#
# TODO : interpolate fvars to macro times. ?mid-points?
#
# TODO : clean-up!
#


def get_output_suffix(iel, ts, naming_scheme, format, output_format):
    if output_format != 'h5':
        if naming_scheme == 'step_iel':
            suffix = '.'.join((ts.suffix % ts.step, format % iel))
        else:
            suffix = '.'.join((format % iel, ts.suffix % ts.step))

    else:
        suffix = format % iel

    return suffix


def convolve_field_scalar(fvars, pvars, iel, ts):
    r"""
    .. math::
      \int_0^t f(t-s) p(s) ds

    Notes
    -----
    - t is given by step
    - f: fvars
      scalar field variables, defined in a micro domain, have shape [step][fmf
      dims]
    - p: pvars
      scalar point variables, a scalar in a point of macro-domain, FMField
      style have shape [n_step][var dims]
    """

    step0 = max(0, ts.step - fvars.steps[-1])

    val = nm.zeros_like(fvars[0])
    for ik in range(step0, ts.step + 1):
        vf = fvars[ts.step-ik]
        vp = pvars[ik][iel, 0, 0, 0]
        val += vf * vp * ts.dt

    return val


def convolve_field_sym_tensor(fvars, pvars, var_name, dim, iel, ts):
    r"""
    .. math::
      \int_0^t f^{ij}(t-s) p_{ij}(s) ds

    Notes
    -----
    - t is given by step
    - f: fvars
      field variables, defined in a micro domain, have shape [step][fmf dims]
    - p: pvars
      sym. tensor point variables, a scalar in a point of
      macro-domain, FMField style, have shape [dim, dim][var_name][n_step][var
      dims]
    """

    step0 = max(0, ts.step - fvars[0, 0][var_name].steps[-1])

    val = nm.zeros_like(fvars[0, 0][var_name][0])
    for ik in range(step0, ts.step + 1):
        for ir in range(dim):
            for ic in range(dim):
                ii = coor_to_sym(ir, ic, dim)
                vf = fvars[ir, ic][var_name][ts.step-ik]
                vp = pvars[ik][iel, 0, ii, 0]
                val += vf * vp * ts.dt
    return val


def add_strain_rs(corrs_rs, strain, vu, dim, iel, out=None):
    if out is None:
        out = nm.zeros_like(corrs_rs[0, 0][vu][0])

    for ir in range(dim):
        for ic in range(dim):
            ii = coor_to_sym(ir, ic, dim)
            out += corrs_rs[ir, ic][vu].data * strain[iel, 0, ii, 0]
    return out


def combine_scalar_grad(corrs, grad, vn, ii, shift_coors=None):
    r"""
    .. math::
      \eta_k \partial_k^x p

    or

    .. math::
      (y_k + \eta_k) \partial_k^x p
    """
    dim = grad.shape[2]

    if shift_coors is None:
        out = corrs[0][vn].data * grad[ii, 0, 0, 0]
        for ir in range(1, dim):
            out += corrs[ir][vn].data * grad[ii, 0, ir, 0]

    else:
        out = (shift_coors[:, 0] + corrs[0][vn].data) * grad[ii, 0, 0, 0]
        for ir in range(1, dim):
            out += (shift_coors[:, ir] + corrs[ir][vn].data) \
                * grad[ii, 0, ir, 0]

    return out


def compute_u_corr_steady(corrs_rs, strain, vu, dim, iel):
    r"""
    .. math::
      \sum_{ij} \bm{\omega}^{ij}\, e_{ij}(\bm{u})

    Notes
    -----
    - iel = element number
    """
    u_corr = add_strain_rs(corrs_rs, strain, vu, dim, iel)
    return u_corr


def compute_u_corr_time(corrs_rs, dstrains, corrs_pressure, pressures,
                        vu, dim, iel, ts):
    r"""
    .. math::
      \sum_{ij} \left[ \int_0^t \bm{\omega}^{ij}(t-s) {\mathrm{d} \over
      \mathrm{d} s} e_{ij}(\bm{u}(s))\,ds\right] + \int_0^t
      \widetilde{\bm{\omega}}^P(t-s)\,p(s)\,ds
    """
    u_corr = convolve_field_scalar(corrs_pressure[vu], pressures,
                                   iel, ts)
    u_corr += convolve_field_sym_tensor(corrs_rs, dstrains, vu,
                                        dim, iel, ts)
    return u_corr


def compute_p_corr_steady(corrs_pressure, pressure, vp, iel):
    r"""
    .. math::
      \widetilde\pi^P\,p
    """
    p_corr = corrs_pressure[vp].data * pressure[iel, 0, 0, 0]
    return p_corr


def compute_p_corr_time(corrs_rs, dstrains, corrs_pressure, pressures,
                        vdp, dim, iel, ts):
    r"""
    .. math::
      \sum_{ij} \int_0^t {\mathrm{d} \over \mathrm{d} t}
      \widetilde\pi^{ij}(t-s)\, {\mathrm{d} \over \mathrm{d} s}
      e_{ij}(\bm{u}(s))\,ds
      + \int_0^t {\mathrm{d} \over \mathrm{d} t}\widetilde\pi^P(t-s)\,p(s)\,ds
    """
    p_corr = convolve_field_scalar(corrs_pressure[vdp], pressures,
                                   iel, ts)
    p_corr += convolve_field_sym_tensor(corrs_rs, dstrains, vdp,
                                        dim, iel, ts)
    return p_corr


def compute_u_from_macro(strain, coor, iel, centre=None):
    r"""
    Macro-induced displacements.

    .. math::
      e_{ij}^x(\bm{u})\,(y_j - y_j^c)
    """
    n_nod, dim = coor.shape

    if centre is None:
        centre = nm.zeros((dim,), dtype=nm.float64)

    n_nod, dim = coor.shape
    um = nm.zeros((n_nod * dim,), dtype=nm.float64)
    for ir in range(dim):
        for ic in range(dim):
            ii = coor_to_sym(ir, ic, dim)
            um[ir::dim] += strain[iel, 0, ii, 0] * (coor[:, ic] - centre[ic])
    return um


def compute_p_from_macro(p_grad, coor, iel, centre=None, extdim=0):
    r"""
    Macro-induced pressure.

    .. math::
      \partial_j^x p\,(y_j - y_j^c)
    """
    n_nod, dim = coor.shape

    if centre is None:
        centre = nm.zeros((dim,), dtype=nm.float64)

    n_nod, dim = coor.shape
    pm = nm.zeros((n_nod,), dtype=nm.float64)
    for ic in range(dim + extdim):
        pm += p_grad[iel, 0, ic, 0] * (coor[:, ic] - centre[ic])
    return pm


def compute_micro_u(corrs, strain, vu, dim, out=None):
    r"""
    Micro displacements.

    .. math::
      \bm{u}^1 = \bm{\chi}^{ij}\, e_{ij}^x(\bm{u}^0)
    """

    if out is None:
        out = nm.zeros_like(corrs[vu+'_00'])

    for ir in range(dim):
        for ic in range(dim):
            ii = coor_to_sym(ir, ic, dim)
            out += corrs[vu+'_%d%d' % (ir, ic)] * strain[ii]
    return out


def compute_stress_strain_u(pb, integral, region, material, vu, data):
    var = pb.create_variables([vu])[vu]
    var.set_data(data)

    stress = pb.evaluate('ev_cauchy_stress.%s.%s(%s, %s)'
                         % (integral, region, material, vu), verbose=False,
                         mode='el_avg', **{vu: var})
    strain = pb.evaluate('ev_cauchy_strain.%s.%s(%s)'
                         % (integral, region, vu), verbose=False,
                         mode='el_avg', **{vu: var})

    return extend_cell_data(stress, pb.domain, region), \
           extend_cell_data(strain, pb.domain, region)


def add_stress_p(out, pb, integral, region, vp, data):
    var = pb.create_variables([vp])[vp]
    var.set_data(data)

    press0 = pb.evaluate('ev_integrate.%s.%s(%s)'
                         % (integral, region, vp), verbose=False,
                         mode='el_avg', **{vp: var})
    press = extend_cell_data(press0, pb.domain, region)

    dim = pb.domain.mesh.dim
    nn = out.shape[0]
    for ii in range(nn):
        for j in range(dim):
            out[ii, 0, j, 0] += press[ii, 0, 0, 0]


def compute_mac_stress_part(pb, integral, region, material, vu, mac_strain):
    avgmat = pb.evaluate('ev_integrate_mat.%s.%s(%s, %s)'
                         % (integral, region, material, vu), verbose=False,
                         mode='el_avg')

    return extend_cell_data(nm.dot(avgmat, mac_strain), pb.domain, region)


def recover_bones(problem, micro_problem, region, eps0,
                  ts, strain, dstrains, p_grad, pressures,
                  corrs_permeability, corrs_rs, corrs_time_rs,
                  corrs_pressure, corrs_time_pressure,
                  var_names, naming_scheme='step_iel'):
    r"""
    Notes
    -----
    - note that

      .. math::
        \widetilde{\pi}^P

      is in corrs_pressure -> from time correctors only 'u', 'dp' are needed.
    """

    dim = problem.domain.mesh.dim

    vu, vp, vn, vpp1, vppp1 = var_names
    vdp = 'd' + vp

    variables = micro_problem.create_variables()
    to_output = variables.create_output

    micro_u, micro_p = variables[vu], variables[vp]

    micro_coor = micro_u.field.get_coor()

    nodes_yc = micro_problem.domain.regions['Yc'].vertices

    join = os.path.join
    format = get_print_info(problem.domain.mesh.n_el, fill='0')[1]

    for ii, iel in enumerate(region.cells):
        print('ii: %d, iel: %d' % (ii, iel))

        pressure = pressures[-1][ii, 0, 0, 0]

        us = corrs_pressure[vu].data * pressure
        add_strain_rs(corrs_rs, strain, vu, dim, ii, out=us)

        ut = convolve_field_scalar(corrs_time_pressure[vu], pressures, ii, ts)
        ut += convolve_field_sym_tensor(corrs_time_rs, dstrains, vu,
                                        dim, ii, ts)
        u1 = us + ut

        u_mic = compute_u_from_macro(strain, micro_coor, ii) + u1

        ps = corrs_pressure[vp].data * pressure

        pt = convolve_field_scalar(corrs_time_pressure[vdp], pressures, ii, ts)
        pt += convolve_field_sym_tensor(corrs_time_rs, dstrains, vdp,
                                        dim, ii, ts)
        p_hat = ps + pt

        # \eta_k \partial_k^x p
        p1 = combine_scalar_grad(corrs_permeability, p_grad, vn, ii)

        p_hat_e = micro_p.field.extend_dofs(p_hat[:, None], fill_value=0.0)
        p_mic = compute_p_from_macro(p_grad, micro_coor, ii)[:, None] \
                + p_hat_e / eps0
        p_mic[nodes_yc] = p1[:, None]

        # (y_k + \eta_k) \partial_k^x p
        p_aux = combine_scalar_grad(corrs_permeability, p_grad, vn, ii,
                                    shift_coors=micro_coor[nodes_yc])

        meval = micro_problem.evaluate

        var_p = variables[vppp1]
        var_p.set_data(p_aux)
        dvel_m1 = meval('ev_diffusion_velocity.i1.Yc(m.K, %s)' % vppp1,
                        verbose=False, mode='el_avg', **{vppp1: var_p})

        var_p = variables[vpp1]
        var_p.set_data(p_hat)
        dvel_m2 = meval('ev_diffusion_velocity.i1.Ym(m.K, %s)' % vpp1,
                        verbose=False, mode='el_avg',
                        **{vpp1: var_p}) * eps0

        out = {}
        out.update(to_output(u_mic, var_info={vu: (True, vu)},
                             extend=True))
        out[vp] = Struct(name='output_data',
                         mode='vertex', data=p_mic,
                         var_name=vp, dofs=micro_p.dofs)

        aux = extend_cell_data(dvel_m1, micro_problem.domain, 'Yc')
        out['dvel_m1'] = Struct(name='output_data',
                                mode='cell', data=aux,
                                dofs=None)

        aux = extend_cell_data(dvel_m2, micro_problem.domain, 'Ym')
        out['dvel_m2'] = Struct(name='output_data',
                                mode='cell', data=aux,
                                dofs=None)

        suffix = get_output_suffix(iel, ts, naming_scheme, format,
                                   micro_problem.output_format)
        micro_name = micro_problem.get_output_name(extra=suffix)
        filename = join(problem.output_dir,
                        'recovered_' + os.path.basename(micro_name))

        micro_problem.save_state(filename, out=out, ts=ts)


def recover_paraflow(problem, micro_problem, region,
                     ts, strain, dstrains, pressures1, pressures2,
                     corrs_rs, corrs_time_rs,
                     corrs_alpha1, corrs_time_alpha1,
                     corrs_alpha2, corrs_time_alpha2,
                     var_names, naming_scheme='step_iel'):

    dim = problem.domain.mesh.dim

    vu, vp = var_names
    vdp = 'd' + vp

    micro_u = micro_problem.variables[vu]
    micro_coor = micro_u.field.get_coor()

    micro_p = micro_problem.variables[vp]

    nodes_y1 = micro_problem.domain.regions['Y1'].vertices
    nodes_y2 = micro_problem.domain.regions['Y2'].vertices

    to_output = micro_problem.variables.create_output

    join = os.path.join
    format = get_print_info(problem.domain.mesh.n_el, fill='0')[1]

    for ii, iel in enumerate(region.cells):
        print('ii: %d, iel: %d' % (ii, iel))

        p1, p2 = pressures1[-1][ii, 0, 0, 0], pressures2[-1][ii, 0, 0, 0]

        us = corrs_alpha1[vu].data * p1 + corrs_alpha2[vu].data * p2
        add_strain_rs(corrs_rs, strain, vu, dim, ii, out=us)

        ut = convolve_field_scalar(corrs_time_alpha1[vu], pressures1, ii, ts)
        ut += convolve_field_scalar(corrs_time_alpha2[vu], pressures2, ii, ts)
        ut += convolve_field_sym_tensor(corrs_time_rs, dstrains, vu,
                                        dim, ii, ts)

        u_corr = us + ut
        u_mic = compute_u_from_macro(strain, micro_coor, ii) + u_corr

        ps = corrs_alpha1[vp].data * p1 + corrs_alpha2[vp].data * p2

        pt = convolve_field_scalar(corrs_time_alpha1[vdp], pressures1, ii, ts)
        pt += convolve_field_scalar(corrs_time_alpha2[vdp], pressures2, ii, ts)
        pt += convolve_field_sym_tensor(corrs_time_rs, dstrains, vdp,
                                        dim, ii, ts)

        p_corr = ps + pt

        p_mic = micro_p.field.extend_dofs(p_corr[:, nm.newaxis])
        p_mic[nodes_y1] = p1
        p_mic[nodes_y2] = p2

        out = {}
        out.update(to_output(u_mic, var_info={vu: (True, vu)}, extend=True))
        out[vp] = Struct(name='output_data',
                         mode='vertex', data=p_mic,
                         var_name=vp, dofs=micro_p.dofs)

        suffix = get_output_suffix(iel, ts, naming_scheme, format,
                                   micro_problem.output_format)
        micro_name = micro_problem.get_output_name(extra=suffix)
        filename = join(problem.output_dir, 'recovered_' + micro_name)

        micro_problem.save_state(filename, out=out, ts=ts)


def save_recovery_region(mac_pb, rname, filename=None):
    filename = get_default(filename, os.path.join(mac_pb.output_dir,
                                                  'recovery_region.vtk'))

    region = mac_pb.domain.regions[rname]

    # Save recovery region characteristic function.
    out = {}
    mask = region.get_charfun(by_cell=False, val_by_id=False)
    out['vmask'] = Struct(name='output_data',
                          mode='vertex', data=mask[:, nm.newaxis],
                          dofs=None)
    mask = region.get_charfun(by_cell=True, val_by_id=False)
    out['cmask'] = Struct(name='output_data',
                          mode='cell',
                          data=mask[:, nm.newaxis, nm.newaxis, nm.newaxis],
                          dofs=None)

    mac_pb.save_state(filename, out=out)


_recovery_global_dict = {}


def destroy_pool():
    if 'pool' in _recovery_global_dict:
        _recovery_global_dict['pool'].close()


atexit.register(destroy_pool)


def _recovery_hook(args):
    idx, label, local_macro, verbose = args
    pb, corrs, recovery_hook = _recovery_global_dict['hook_args']

    output.set_output(quiet=False)
    output.level = label[1]
    output(label[0])
    output.set_output(quiet=not(verbose))

    if isinstance(corrs, list):
        corrs = corrs[idx]

    return recovery_hook(pb, corrs, local_macro)


def get_recovery_points(region, eps0):
    rcoors = region.domain.mesh.coors[region.get_entities(0), :]
    rcmin = nm.min(rcoors, axis=0)
    rcmax = nm.max(rcoors, axis=0)
    nn = nm.round((rcmax - rcmin) / eps0)
    if nm.prod(nn) == 0:
        raise ValueError(
            'incompatible recovery region and microstructure size!')

    cs = []
    for ii, n in enumerate(nn):
        cs.append(nm.arange(n) * eps0 + rcmin[ii])

    ccoors = nm.empty((int(nm.prod(nn)), nn.shape[0]), dtype=nm.float64)
    for ii, icoor in enumerate(nm.meshgrid(*cs, indexing='ij')):
        ccoors[:, ii] = icoor.flatten()

    return ccoors


def recover_micro_hook(micro_filename, region, macro, eps0,
                       region_mode='el_centers', eval_mode='constant',
                       eval_vars=None, corrs=None, recovery_file_tag='',
                       define_args=None, output_dir=None, verbose=False):
    """
    Parameters
    ----------
    micro_filename : str
        The definition file of the microproblem.
    region : Region or array
        The macroscopic region to be recovered. If array, the centers of
        microscopic RVEs (Representative Volume Element) are assumed to be
        stored in it. If Region, the RVE centers are computed according to
        `region_mode`, see below.
    macro : dict of arrays or tuples
        Either macroscopic values (if array) or the tuple
        (mode, eval_var, nodal_values) is expected. The tuple is used to
        evaluate the macroscopic values in given points of RVEs (see
        'eval_mode`). `mode` can be 'val', 'grad', 'div', or 'cauchy_strain'.
    eps0 : float
        The size of the microstructures (RVE).
    region_mode : {'el_centers', 'tiled'}
        If 'el_centers', the RVE centers are identical to the element centers
        of the macroscopic FE mesh. If 'tiled', the recovered region is tiled
        by rescaled RVEs.
    eval_mode : {'constant', 'continuous'}
        If 'constant', the macroscopic fields are evaluated only at the RVE
        centers. If 'continuous', the fields are evaluated at all points of
        the RVE mesh.
    eval_vars : list of variables
        The list of variables use to evaluate the macroscopic fields.
    corrs : dict of CorrSolution
        The correctors for recovery.
    recovery_file_tag : str
        The tag which is appended to the output file.
    define_args : dict
        The define arguments for the microscopic problem.
    output_dir : str
        The output directory.
    verbose : bool
        The verbose terminal output.
    """
    import sfepy.base.multiproc_proc as multi

    if 'micro_problem' not in _recovery_global_dict:
        # Create a micro-problem instance.
        required, other = get_standard_keywords()
        required.remove('equations')
        conf = ProblemConf.from_file(micro_filename, required, other,
                                     verbose=False, define_args=define_args)
        if output_dir is not None:
            conf.options.output_dir = output_dir

        recovery_hook = conf.options.get('recovery_hook', None)
        pb = None
        if recovery_hook is not None:
            recovery_hook = conf.get_function(recovery_hook)

            if corrs is None:
                # Coefficients and correctors
                coefs_filename = conf.options.get('coefs_filename', 'coefs')
                coefs_filename = op.join(conf.options.get('output_dir', '.'),
                                         coefs_filename + '.h5')
                coefs = Coefficients.from_file_hdf5(coefs_filename)
                corrs = \
                    get_correctors_from_file_hdf5(dump_names=coefs.save_names)

            pb = Problem.from_conf(conf, init_equations=False,
                                   init_solvers=False)

        _recovery_global_dict['micro_problem'] = pb, corrs, recovery_hook
    else:
        pb, corrs, recovery_hook = _recovery_global_dict['micro_problem']

    is_multiproc = pb.conf.options.get('multiprocessing', True)\
        and multi.use_multiprocessing

    if recovery_hook is not None:
        if eval_mode not in ['constant', 'continuous']:
            raise ValueError(f"Evaluation mode '{eval_mode}' not implemented!")

        if isinstance(region, Region):
            if region_mode == 'el_centers':
                dim = region.domain.mesh.dim
                ccoors = region.cmesh.get_centroids(dim)
                ccoors = ccoors[region.get_cells(), :]
                recovery_ids = region.get_entities(-1)
                recovery_ids_s = [f'(element {k})' for k in recovery_ids]
            elif region_mode == 'tiled':
                ccoors = get_recovery_points(region, eps0)
                recovery_ids = nm.arange(ccoors.shape[0])
                recovery_ids_s = [f'(point {k})' for k in recovery_ids]
            else:
                raise ValueError(
                    f"Region mode '{region_mode}' not implemented!")

        else:
            ccoors = region
            recovery_ids = nm.arange(ccoors.shape[0])
            recovery_ids_s = [f'(point {k})' for k in recovery_ids]

        nrec = ccoors.shape[0]

        output(f'recovering {nrec} microsctructures...')
        timer = Timer(start=True)
        output_level, output_fun = output.level, output.output_function

        _recovery_global_dict['hook_args'] = pb, corrs, recovery_hook

        if is_multiproc:
            multi_local_macro = []
            num_workers = nm.min([multi.cpu_count(), nrec])
            _recovery_global_dict['pool'] = multi.Pool(processes=num_workers)
        else:
            outs = []

        # Recover micro. region
        mesh = pb.domain.mesh
        bbox = mesh.get_bounding_box()
        mic_coors = (mesh.coors - 0.5 * (bbox[1, :] + bbox[0, :])) * eps0
        coors, conn, ndoffset, rec_ids = [], [], 0, []

        if eval_vars is not None and eval_mode == 'constant':
            new_macro = {}
            for k, v in macro.items():
                if isinstance(v, tuple):
                    emode, evar, val = v
                    efield = eval_vars[evar].field
                    new_macro[k] = efield.evaluate_at(ccoors, val, mode=emode)
                else:
                    new_macro[k] = v

            macro = new_macro

        for ii in range(nrec):
            local_coors = mic_coors.copy()
            local_coors[:, :ccoors.shape[1]] += ccoors[ii]
            coors.append(local_coors)
            conn.append(mesh.get_conn(mesh.descs[0]) + ndoffset)
            ndoffset += mesh.n_nod
            rec_ids.append(nm.ones((mesh.n_el,)) * recovery_ids[ii])

            label = (f'micro: {ii + 1}/{nrec} {recovery_ids_s[ii]}',
                     output_level)

            local_macro = {'eps0': eps0}

            if eval_vars is not None and eval_mode == 'continuous':
                for k, v in macro.items():
                    if isinstance(v, tuple):
                        emode, evar, val = v
                        efield = eval_vars[evar].field

                        # # Inside recovery region?
                        # from sfepy.base.base import debug; debug()
                        # v = nm.ones((efield.region.entities[0].shape[0], 1))
                        # v[evfield.vertex_remap[region.entities[0]]] = 0
                        # no = nm.sum(v)
                        # aux = efield.evaluate_at(local_coors, v)
                        # if no > 0 and (nm.sum(aux) / no) > 1e-3:
                        #     print('not inside!')
                        #     continue

                        local_macro[k] = efield.evaluate_at(local_coors, val,
                                                            mode=emode)
                    else:
                        local_macro[k] = v
            else:
                for k, v in macro.items():
                    if k.startswith('_'):
                        local_macro[k] = v
                    else:
                        local_macro[k] = v[ii, ...]

            if is_multiproc:
                multi_local_macro.append((ii, label, local_macro, verbose))
            else:
                outs.append(_recovery_hook((ii, label, local_macro, verbose)))

        rec_ids = nm.hstack(rec_ids).reshape((-1, 1, 1, 1))

        if is_multiproc:
            pool = _recovery_global_dict['pool']
            outs = list(pool.map(_recovery_hook, multi_local_macro))

        output.level, output.output_function = output_level, output_fun

        # Collect output data - by region
        outregs_data = {}
        outregs_info = {None: ('ALL', nm.arange(pb.domain.mesh.n_el))}
        vn_reg = {None: None}
        for k, v in outs[0].items():
            if hasattr(v, 'region_name'):
                rn = v.region_name
                if rn not in outregs_info:
                    reg = pb.domain.regions[rn]
                    outregs_info[rn] = (rn, reg.get_entities(-1))
            else:
                vn = getattr(v, 'var_name', None)
                if vn not in vn_reg:
                    reg = pb.create_variables(vn)[vn].field.region
                    rn = reg.name
                    vn_reg[vn] = rn
                    if rn not in outregs_info:
                        outregs_info[rn] = (rn, reg.get_entities(-1))
                else:
                    rn = vn_reg[vn]

            if rn in outregs_data:
                outregs_data[rn].append(k)
            else:
                outregs_data[rn] = [k]

        nrve = len(coors)
        coors = nm.vstack(coors)
        ngroups = nm.tile(mesh.cmesh.vertex_groups.squeeze(), (nrve,))
        conn = nm.vstack(conn)
        cgroups = nm.tile(mesh.cmesh.cell_groups.squeeze(), (nrve,))
        # Get region mesh and data
        output_dir = pb.conf.options.get('output_dir', '.')
        for rn in outregs_data.keys():
            rlabel, cidxs = outregs_info[rn]
            gcidxs = nm.hstack([cidxs + mesh.n_el * ii for ii in range(nrve)])
            rconn = conn[gcidxs]
            remap = -nm.ones((coors.shape[0],), dtype=nm.int32)
            remap[rconn] = 1
            vidxs = nm.where(remap > 0)[0]
            remap[vidxs] = nm.arange(len(vidxs))
            rconn = remap[rconn]
            rcoors = coors[vidxs, :]

            out = {}
            for ivar in outregs_data[rn]:
                data = [outs[ii][ivar].data for ii in range(nrve)]
                out[ivar] = Struct(name='output_data',
                                   mode=outs[0][ivar].mode,
                                   data=nm.vstack(data))

            out['recovery_id'] = Struct(name='output_data',
                                        mode='cell',
                                        data=rec_ids[gcidxs, ...])
            micro_name = pb.get_output_name(extra='recovered%s_%s'
                                            % (recovery_file_tag, rlabel))
            filename = op.join(output_dir, op.basename(micro_name))
            mesh_out = Mesh.from_data('recovery_%s' % rlabel, rcoors,
                                      ngroups[vidxs], [rconn],
                                      [cgroups[gcidxs]], [mesh.descs[0]])
            mesh_out.write(filename, io='auto', out=out)
            output(f'output saved to "{filename}"')

        output('...done in %.2f s' % timer.stop())



"""
Elapsed time measurement utilities.
"""
import time

from sfepy.base.base import Struct, IndexedStruct

class Timer(Struct):

    def __init__(self, name='timer', start=False):
        Struct.__init__(self, name=name)
        self.time_function = time.perf_counter
        self.reset()

        if start:
            self.start()

    def reset(self):
        self.t0 = self.t1 = None
        self.total = self.dt = 0.0

    def start(self, reset=False):
        if reset: self.reset()

        self.t1 = None
        self.t0 = self.time_function()

    def stop(self):
        self.t1 = self.time_function()
        if self.t0 is None:
            raise ValueError('timer "%s" was not started!' % self.name)

        self.dt = self.t1 - self.t0
        self.total += self.dt
        return self.dt

    def add(self, dt):
        if self.t1 is None:
            raise ValueError('timer "%s" was not started and stopped!'
                             % self.name)

        self.t1 += dt
        self.dt += dt
        self.total += dt

class Timers(IndexedStruct):

    def __init__(self, names):
        IndexedStruct.__init__(self,
                               **{name : Timer(name=name) for name in names})

    def create(self, name):
        if name in self.__dict__:
            self[name].reset()

        else:
            self[name] = Timer(name=name)

    def start(self, name):
        self[name].start()

    def stop(self, name):
        return self[name].stop()

    def get_dts(self):
        return {name : timer.dt for name, timer in self.to_dict().items()}

    def get_totals(self):
        return {name : timer.total for name, timer in self.to_dict().items()}



"""
Time step controllers.
"""
import numpy as nm

from sfepy.base.base import Struct
from sfepy.solvers.solvers import TimeStepController

class FixedTSC(TimeStepController):
    """
    Fixed (do-nothing) time step controller.
    """
    name = 'tsc.fixed'

class TimesSequenceTSC(TimeStepController):
    """
    Given times sequence time step controller.
    """
    name = 'tsc.time_sequence'

    _parameters = [
        ('times', 'iterable', range(1, 6), True,
         'A sequence of times to generate.'),
    ]

    def __init__(self, conf, **kwargs):
        TimeStepController.__init__(self, conf=conf, **kwargs)

        self.iter_times = iter(self.conf.times)

    def get_initial_dt(self, ts, vec, **kwargs):
        # This cannot be called repeatedly!
        self.t0 = ts.t0
        return next(self.iter_times) - self.t0

    def __call__(self, ts, vec0, vec1, **kwargs):
        self.t0 = ts.time
        try:
            t1 = next(self.iter_times)

        except StopIteration:
            t1 = ts.t1

        new_dt = t1 - self.t0
        if new_dt == 0.0:
            new_dt = 1.0

        status = Struct(u_err=None, v_err=None, emax=None, result='accept')

        return new_dt, status

def eval_scaled_norm(terr, eps_a, eps_r):
    return nm.sqrt(
        1 / len(terr) *
        nm.sum((terr / (eps_r * nm.abs(terr) + eps_a))**2)
    )

class ElastodynamicsBasicTSC(TimeStepController):
    """
    Adaptive time step I-controller for elastodynamics.

    The implementation is based on [1]. The default parameters correspond to
    the PID-Controller as implemented in ``tsc.ed_pid`` with P=D=0, I=1.

    [1] Grafenhorst, Matthias, Joachim Rang, and Stefan Hartmann.
    “Time-Adaptive Finite Element Simulations of Dynamical Problems for
    Temperature-Dependent Materials.” Journal of Mechanics of Materials and
    Structures 12, no. 1 (November 26, 2016): 57–91.
    https://doi.org/10.2140/jomms.2017.12.57.
    """
    name = 'tsc.ed_basic'

    _parameters = [
        ('eps_r', 'list of floats or float', None, True,
         'Relative tolerance(s).'),
        ('eps_a', 'list of floats or float', None, True,
         'Absolute tolerance(s).'),
        ('fmin', 'float', 0.3, False,
         'Minimum step size change factor on step rejection.'),
        ('fmax', 'float', 2.5, False,
         'Maximum step size change factor on step acceptance.'),
        ('fsafety', 'float', 0.8, False,
         'Step size change safety factor.'),
        ('error_order', 'float', 2, False,
         'The order of the solver error estimate.'),
        ('guess_dt0', 'bool', False, False,
         'Guess a good initial step size from initial conditions.'),
    ]

    @staticmethod
    def get_scaled_errors(dt, vec0, vec1, eps_as, eps_rs, unpack):
        u_eps_a, v_eps_a = eps_as
        u_eps_r, v_eps_r = eps_rs
        aux = unpack(vec0)
        u0, v0, a0 = aux if unpack.n_arg == 3 else (aux[0], aux[2], aux[3])
        aux = unpack(vec1)
        u1, v1, a1 = aux if unpack.n_arg == 3 else (aux[0], aux[2], aux[3])

        # Backward Euler step.
        u1_be = u0 + dt * v1
        v1_be = v0 + dt * a1
        # Truncation error estimates.
        u_terr = u1 - u1_be
        v_terr = v1 - v1_be

        u_err = eval_scaled_norm(u_terr, u_eps_a, u_eps_r)
        v_err = eval_scaled_norm(v_terr, v_eps_a, v_eps_r)

        return u_err, v_err

    def get_initial_dt(self, ts, vec, unpack, **kwargs):
        """
        Adapted from [1] for second order ODEs.

        [1] Hairer, Ernst, Gerhard Wanner, and Syvert P. Nørsett. Solving
        Ordinary Differential Equations I: Nonstiff Problems. Vol. 8. Springer
        Series in Computational Mathematics. Berlin, Heidelberg: Springer,
        1993. https://doi.org/10.1007/978-3-540-78862-1.
        """
        conf = self.conf
        if not conf.guess_dt0:
            return ts.dt

        error_order = conf.error_order
        eps_a = min(*conf.eps_a)
        eps_r = min(*conf.eps_r)

        aux = unpack(vec0)
        u0, v0, a0 = aux if unpack.n_arg == 3 else (aux[0], aux[2], aux[3])

        d0 = eval_scaled_norm(u0, eps_a, eps_r)
        d1 = eval_scaled_norm(v0, eps_a, eps_r)
        d2 = eval_scaled_norm(a0, eps_a, eps_r)
        md1d2 = max(d1, d2)

        h0 = 1e-6 if (d0 < 1e-5) or (d1 < 1e-5) else 0.01 * (d0 / d1)
        h1 = (max(1e-6, h0 * 1e-3) if md1d2 < 1e-15
              else (0.01 / md1d2) ** (1 / error_order))

        dt0 = min(100 * h0, h1)
        return dt0

    def __call__(self, ts, vec0, vec1, unpack, **kwargs):
        conf = self.conf
        dt = ts.dt
        fmin, fmax, fsafety, error_order = (
            conf.fmin, conf.fmax, conf.fsafety, conf.error_order,
        )
        aux = unpack(vec0)
        u0, v0, a0 = aux if unpack.n_arg == 3 else (aux[0], aux[2], aux[3])
        aux = unpack(vec1)
        u1, v1, a1 = aux if unpack.n_arg == 3 else (aux[0], aux[2], aux[3])

        u_err, v_err = self.get_scaled_errors(
            dt, vec0, vec1, conf.eps_a, conf.eps_r, unpack,
        )
        emax = max(u_err, v_err)
        status = Struct(u_err=u_err, v_err=v_err, emax=emax)
        if emax <= 1:
            # Step accepted.
            new_dt = dt * min(fmax, fsafety / emax**(1.0 / error_order))
            status.result = 'accept'

        else:
            new_dt = dt * max(fmin, fsafety / emax**(1.0 / error_order))
            status.result = 'reject'

        return new_dt, status

class ElastodynamicsPIDTSC(ElastodynamicsBasicTSC):
    """
    Adaptive time step PID controller for elastodynamics.

    The implementation is based on [1], [2] (PI Controller) and [3] (PID).
    The default parameters correspond to the I-Controller as implemented in
    ``tsc.ed_basic``.

    [1] Grafenhorst, Matthias, Joachim Rang, and Stefan Hartmann.
    “Time-Adaptive Finite Element Simulations of Dynamical Problems for
    Temperature-Dependent Materials.” Journal of Mechanics of Materials and
    Structures 12, no. 1 (November 26, 2016): 57–91.
    https://doi.org/10.2140/jomms.2017.12.57.
    [2] Hairer, Ernst, Syvert Paul Nørsett, and Gerhard Wanner. Solving
    Ordinary Differential Equations II: Stiff and Differential-Algebraic
    Problems. Springer Science & Business Media, 1993.
    [3] Söderlind, Gustaf. “Digital Filters in Adaptive Time-Stepping.” ACM
    Transactions on Mathematical Software 29, no. 1 (March 1, 2003): 1–26.
    https://doi.org/10.1145/641876.641877.
    """
    name = 'tsc.ed_pid'

    _parameters = ElastodynamicsBasicTSC._parameters + [
        ('pcoef', 'float', 0.0, False,
         'Proportional (P) coefficient of the step size control.'),
        ('icoef', 'float', 1.0, False,
         'Intregral (I) coefficient of the step size control.'),
        ('dcoef', 'float', 0.0, False,
         'Derivative (D) coefficient of the step size control.'),
    ]

    def __init__(self, conf, **kwargs):
        ElastodynamicsBasicTSC.__init__(self, conf=conf, **kwargs)

        self.emax0 = 1.0
        self.emax00 = 1.0

    def __call__(self, ts, vec0, vec1, unpack, **kwargs):
        conf = self.conf
        dt = ts.dt
        fmin, fmax, fsafety, pcoef, icoef, dcoef, error_order = (
            conf.fmin, conf.fmax, conf.fsafety,
            conf.pcoef, conf.icoef, conf.dcoef,
            conf.error_order,
        )
        b1 = -(pcoef + icoef + dcoef) / error_order
        b2 = (pcoef + 2 * dcoef) / error_order
        b3 = -dcoef / error_order
        u_err, v_err = self.get_scaled_errors(
            dt, vec0, vec1, conf.eps_a, conf.eps_r, unpack,
        )
        emax = max(u_err, v_err)

        c1 = 1 if b1 == 0 else emax**b1
        c2 = 1 if b2 == 0 else self.emax0**b2
        c3 = 1 if b3 == 0 else self.emax00**b3
        new_dt = dt * min(fmax, max(fmin, fsafety * c1 * c2 * c3))

        status = Struct(u_err=u_err, v_err=v_err, emax=emax)
        status.result = 'accept' if emax <= 1 else 'reject'

        self.emax00 = self.emax0
        self.emax0 = emax

        return new_dt, status

class ElastodynamicsLinearTSC(ElastodynamicsBasicTSC):
    """
    Adaptive time step controller for elastodynamics and linear problems.

    Simple heuristics around :class:`ElastodynamicsBasicTSC` that increases the
    step size only after a sufficient number of accepted iterations passed and
    the increase is large enough. In particular:

    - Let new_dt be the step size proposed by `tsc.ed_basic` and dt the current
      step size.
    - If the current step is rejected, the count attribute is reset to zero and
      fred * new_dt is returned.
    - If the current step is accepted:

      - If the count is lower than inc_wait, it is incremented and dt is
        returned.
      - Otherwise, if (new_dt / dt) >= min_finc (>= 1), the count
        is reset to zero and new_dt is returned.
      - Else, if (new_dt / dt) < min_finc, dt is returned.
    """
    name = 'tsc.ed_linear'

    _parameters = ElastodynamicsBasicTSC._parameters + [
        ('fred', 'float', 1.0, False,
         'Additional step size reduction factor w.r.t. `tsc.ed_basic`.'),
        ('inc_wait', 'int', 10, False,
         """The number of consecutive accepted steps to wait before increasing
            the step size."""),
        ('min_finc', 'float >= 1', 1.5, False,
         'Minimum step size increase factor.'),
    ]

    def __init__(self, conf, **kwargs):
        ElastodynamicsBasicTSC.__init__(self, conf=conf, **kwargs)

        if self.conf.min_finc < 1.0:
            raise ValueError(
                f'min_finc must be >= 1! (is {conf.min_finc})'
            )
        self.count = 0

    def __call__(self, ts, vec0, vec1, unpack, **kwargs):
        conf = self.conf
        dt = ts.dt

        new_dt, status = ElastodynamicsBasicTSC.__call__(
            self, ts, vec0, vec1, unpack, **kwargs
        )
        if status.result == 'reject':
            self.count = 0
            new_dt *= conf.fred

        else: # accept
            if self.count >= conf.inc_wait:
                if (new_dt / dt) >= conf.min_finc:
                    self.count = 0

                else:
                    new_dt = dt

            else:
                self.count += 1
                new_dt = dt

        return new_dt, status



import numpy as nm

import sfepy.base.testing as tst

expected_normals = { # Need to be normalized!
    '2_3' : nm.array([[ 0, -1],
                      [ 1,  1],
                      [-1,  0]], dtype=nm.float64),
    '2_4' : nm.array([[ 0, -1],
                      [ 1,  0],
                      [ 0,  1],
                      [-1,  0]], dtype=nm.float64),
    '3_4' : nm.array([[ 0,  0, -1],
                      [-1,  0,  0],
                      [ 0, -1,  0],
                      [ 1,  1,  1]], dtype=nm.float64),
    '3_8' : nm.array([[ 0,  0, -1],
                      [-1,  0,  0],
                      [ 0, -1,  0],
                      [ 0,  0,  1],
                      [ 1,  0,  0],
                      [ 0,  1,  0]], dtype=nm.float64),
}

def test_normals():
    """
    Check orientations of surface normals on the reference elements.
    """
    import sfepy
    from sfepy.discrete import Integral, PolySpace
    from sfepy.discrete.fem import Mesh, FEDomain
    from sfepy.discrete.fem.mappings import FEMapping
    from sfepy.linalg import normalize_vectors

    ok = True

    for geom in ['2_3', '2_4', '3_4', '3_8']:
        mesh = Mesh.from_file('meshes/elements/%s_1.mesh' % geom,
                              prefix_dir=sfepy.data_dir)
        domain = FEDomain('domain', mesh)
        surface = domain.create_region('Surface', 'vertices of surface',
                                       'facet')
        domain.create_surface_group(surface)

        sd = domain.surface_groups[surface.name]

        coors = domain.get_mesh_coors()
        gel = domain.geom_els[geom].surface_facet
        ps = PolySpace.any_from_args('aux', gel, 1)

        mapping = FEMapping(coors, sd.get_connectivity(), ps)

        integral = Integral('i', order=1)
        vals, weights = integral.get_qp(gel.name)

        # Evaluate just in the first quadrature point...
        geo = mapping.get_mapping(vals[:1], weights[:1], is_face=True)

        expected = expected_normals[geom].copy()
        normalize_vectors(expected)

        _ok = nm.allclose(expected, geo.normal[:, 0, :, 0],
                          rtol=0.0, atol=1e-14)
        tst.report('%s: %s' % (geom, _ok))

        if not _ok:
            tst.report('expected:')
            tst.report(expected)
            tst.report('actual:')
            tst.report(geo.normal[:, 0, :, 0])

        ok = ok and _ok

    assert ok



"""
Finite element reference mappings.
"""
import numpy as nm

from sfepy import Config
from sfepy.base.base import get_default, output
from sfepy.base.mem_usage import raise_if_too_large
from sfepy.discrete.common.mappings import Mapping, PyCMapping
from sfepy.discrete import PolySpace
from sfepy.linalg.utils import invs_fast, dets_fast
from sfepy.linalg import dot_sequences


def tranform_coors_to_lower_dim(coors, to_dim):
    """
    Transform element coordinates into XY plane.

    See:
      https://math.stackexchange.com/questions/1167717/transform-a-plane-to-the-xy-plane
      https://en.wikipedia.org/wiki/Rotation_matrix#Rotation_matrix_from_axis_and_angle
    """
    if coors.shape[-1] == 3 and to_dim == 2:
        bv = nm.cross(coors[:, 1, :] - coors[:, 0, :],
                      coors[:, -1, :] - coors[:, 0, :])

        b = nm.linalg.norm(bv, axis=1)
        ab = nm.sqrt(bv[:, 0]**2 + bv[:, 1]**2)
        cphi = bv[:, 2] / b
        sphi = ab / b

        mtxR = nm.eye(3, 3) * cphi[:, None, None]
        idxs = nm.where(nm.abs(ab) > 1e-16)[0]

        if len(idxs) > 0:
            u1 = bv[idxs, 1] / ab[idxs]
            u2 = - bv[idxs, 0] / ab[idxs]
            cphi1 = 1 - cphi[idxs]
            sphi = sphi[idxs]

            R0 = nm.array([
                [u1**2 * cphi1, u1 * u2 * cphi1, u2 * sphi],
                [u1 * u2 * cphi1, u2**2 * cphi1, -u1 * sphi],
                [-u2 * sphi, u1 * sphi, 0]
            ]).transpose(2, 0, 1)

            mtxR[idxs, ...] += R0

        out = nm.einsum('qij,qkj->qki', mtxR, coors, optimize=True)

        return out[:, :, :to_dim]
    else:
        msg = (f'Coordinate transformation from dimension {coors.shape[-1]}'
               f' to dimension {to_dim} is not supported!')
        raise NotImplemented(msg)


def eval_mapping_data_in_qp(coors, conn, bf_g, weights,
                            ebf_g=None, is_face=False, eps=1e-15,
                            se_conn=None, se_bf_bg=None, ecoors=None):
    """
    Evaluate mapping data.

    Parameters
    ----------
    coors: numpy.ndarray
        The nodal coordinates.
    conn: numpy.ndarray
        The element connectivity.
    bf_g: numpy.ndarray
        The derivatives of the domain basis functions with respect to the
        reference coordinates.
    weights: numpy.ndarray
        The weights of the quadrature points.
    ebf_g: numpy.ndarray
        The derivatives of the field basis functions with respect to the
        reference coordinates.
    is_face: bool
        Is it the boundary of a region?
    eps: float
        The tolerance for the normal vectors calculation.
    se_conn: numpy.ndarray
        The connectivity for the calculation of surface derivatives.
    se_bf_bg: numpy.ndarray
        The surface basis function derivatives with respect to the reference
        coordinates.
    ecoors: numpy.ndarray
        The element nodal coordinates.

    Returns
    -------
    det: numpy.ndarray
        The determinant of the mapping evaluated in integration points.
    volume: numpy.ndarray
        The element (volume or surface) volumes in integration points.
    bfg: numpy.ndarray
        The derivatives of the basis functions with respect to the spatial
        coordinates. Can be evaluated either for surface elements if `bf_g`,
        `se_conn`, and `se_bf_bg` are given.
    normal: numpy.ndarray
        The normal vectors for the surface elements in integration points.
    """
    if ecoors is None:
        ecoors = coors[conn, :]

    if bf_g.ndim == 4:
        mtxRM = nm.einsum('cqij,cjk->cqik', bf_g, ecoors, optimize=True)
    else:
        mtxRM = nm.einsum('qij,cjk->cqik', bf_g, ecoors, optimize=True)

    n_el, n_qp = mtxRM.shape[:2]

    if is_face:
        # outward unit normal vector
        dim = coors.shape[1]
        normal = nm.ones((n_el, n_qp, dim, 1), dtype=nm.float64)
        if dim == 1:
            det = nm.tile(weights, (n_el, 1)).reshape(n_el, n_qp, 1, 1)
            ii = nm.where(conn == 0)[0]
            normal[ii] *= -1.0
        elif dim == 2:
            c1, c2 = mtxRM[..., 0], mtxRM[..., 1]
            det0 = nm.sqrt(c1**2 + c2**2).reshape(n_el, n_qp, 1, 1)
            det = det0 * weights[..., None, None]
            det0[nm.abs(det0) < eps] = 1.0
            normal[..., 0, :] = c2
            normal[..., 1, :] = -c1
            normal /= det0
        elif dim == 3:
            j012 = mtxRM[..., 0, :].T
            j345 = mtxRM[..., 1, :].T
            c1 = (j012[1] * j345[2] - j345[1] * j012[2]).T
            c2 = (j012[0] * j345[2] - j345[0] * j012[2]).T
            c3 = (j012[0] * j345[1] - j345[0] * j012[1]).T
            det0 = nm.sqrt(c1**2 + c2**2 + c3**2).reshape(n_el, n_qp, 1, 1)
            det = det0 * weights[..., None, None]
            det0[nm.abs(det0) < eps] = 1.0
            normal[..., 0, 0] = c1
            normal[..., 1, 0] = -c2
            normal[..., 2, 0] = c3
            normal /= det0
    else:
        # det0 = nm.linalg.det(mtxRM)
        det0 = dets_fast(mtxRM)
        if nm.any(det0 <= 0.0):
            raise ValueError('warp violation!')

        det = det0 * weights
        det = det.reshape(n_el, n_qp, 1, 1)
        normal = None

    if ebf_g is not None:
        if is_face and se_conn is not None and se_bf_bg is not None:
            mtxRM = nm.einsum('cqij,cjk->cqik', se_bf_bg, coors[se_conn, :],
                              optimize=True)
            mtxRMI = invs_fast(mtxRM)
            bfg = nm.einsum('cqij,cqjk->cqik', mtxRMI, ebf_g, optimize=True)
        else:
            mtxRMI = invs_fast(mtxRM, det0)
            es_arg2 = 'x' if ebf_g.shape[0] == 1 else 'c'
            bfg = nm.einsum(f'cqij,{es_arg2}qjk->cqik', mtxRMI, ebf_g,
                            optimize=True)
    else:
        bfg = None

    volume = nm.sum(det, axis=1).reshape((n_el, 1, 1, 1))

    return det, volume, bfg, normal


class FEMapping(Mapping):
    """
    Base class for finite element mappings.
    """

    def __init__(self, coors, conn, poly_space=None, gel=None, order=1):
        self.coors = coors
        self.conn = conn

        try:
            nm.take(self.coors, self.conn)

        except IndexError:
            output('coordinates shape: %s' % list(coors.shape))
            output('connectivity: min: %d, max: %d' % (conn.min(), conn.max()))
            msg = 'incompatible connectivity and coordinates (see above)'
            raise IndexError(msg)

        self.n_el, self.n_ep = conn.shape
        self.dim = self.coors.shape[1]

        if isinstance(gel, dict):
            poly_space = {}
            for k, v in gel.items():
                poly_space[k] = PolySpace.any_from_args(None, v, order,
                                                        base='lagrange',
                                                        force_bubble=False)
        elif poly_space is None:
            poly_space = PolySpace.any_from_args(None, gel, order,
                                                 base='lagrange',
                                                 force_bubble=False)

        self.poly_space = poly_space
        self.indices = None

    def get_geometry(self):
        """
        Return reference element geometry as a GeometryElement instance.
        """
        if isinstance(self.poly_space, dict):
            return {k: v.geometry for k, v in self.poly_space.items()}
        else:
            return self.poly_space.geometry

    def get_base(self, coors, diff=False, grad_axes=None):
        """
        Get basis functions or their gradient evaluated in given
        coordinates.
        """
        if isinstance(self.poly_space, dict):
            bf = {k: v.eval_base(coors[k], diff=diff)
                  for k, v in self.poly_space.items()}
        else:
            bf = self.poly_space.eval_base(coors, diff=diff)

        indices = self.indices
        if indices is not None:
            ii = max(self.dim - 1, 1)
            if isinstance(indices, dict):
                # Treat elements with different facet types, e.g. wedges
                n = max(len(v) for v in indices.values())
                bf_ = nm.zeros((len(indices), bf[0].shape[0], ii, n),
                               dtype=nm.float64)
                if diff and grad_axes is not None:
                    for k, v in indices.items():
                        bf_[k, ..., :v.shape[0]] =\
                            bf[k][..., grad_axes[k], :][..., v]
                else:
                    for k, v in indices.items():
                        bf_[k, ..., :v.shape[0]] = bf[k][..., :ii:, v]

                bf = bf_
            else:
                bf = nm.ascontiguousarray(bf[..., :ii:, indices])

        return bf

    def set_basis_indices(self, indices):
        """
        Set indices to cell-based basis that give the facet-based basis.
        """
        self.indices = indices

    def get_physical_qps(self, qp_coors):
        """
        Get physical quadrature points corresponding to given reference
        element quadrature points.

        Returns
        -------
        qps : array
            The physical quadrature points ordered element by element,
            i.e. with shape (n_el, n_qp, dim).
        """
        bf = self.get_base(qp_coors)
        if isinstance(bf, dict):
            sh = self.conn.shape
            nqp = max(v.shape[0] for v in bf.values())
            bf_ = nm.zeros((sh[0], nqp, 1, sh[1]), dtype=nm.float64)
            ft = nm.count_nonzero(nm.diff(nm.sort(self.conn)), axis=1) + 1
            for k, v in bf.items():
                idxs = ft == k
                if nm.any(idxs):
                    bf_[idxs, :v.shape[0], :, :k] = v

            qps = dot_sequences(bf_[..., 0, :], self.coors[self.conn])
        else:
            qps = nm.dot(nm.atleast_2d(bf.squeeze()), self.coors[self.conn])
            # Reorder so that qps are really element by element.
            qps = nm.ascontiguousarray(nm.swapaxes(qps, 0, 1))

        return qps

    def get_mapping(self, qp_coors, weights, bf=None, poly_space=None,
                    ori=None, transform=None, is_face=False, fc_bf_map=None,
                    extra=(None, None, None)):
        """
        Get the mapping for given quadrature points, weights, and
        polynomial space.

        Parameters
        ----------
        qp_coors: numpy.ndarray
            The coordinates of the integration points.
        weights:
            The integration weights.
        bf: numpy.ndarray
            The basis functions.
        poly_space: PolySpace instance
            The PolySpace instance.
        ori: numpy.ndarray
            Element orientation, used by hierarchical basis.
        transform: numpy.ndarray
            The transformation matrix applied to the basis functions.
        is_face: bool
            Is it the boundary of a region?
        fc_bf_map: tuple
            The additional info to remap face derivatives of wedge elements:
            - fc_bf_map[0]: id of face group (triangle or quad)
            - fc_bf_map[1]: position of inplane derivatives (xy-axes)
        extra: tuple
            The extra data for surface derivatives:
            - the derivatives of the field boundary basis functions with
              respect to the reference coordinates
            - the boundary connectivity
            - the derivatives of the domain boundary basis functions with
              respect to the reference coordinates

        Returns
        -------
        pycmap: PyCMapping instance
            The domain mapping data.
        """
        poly_space = get_default(poly_space, self.poly_space)

        if fc_bf_map is not None:
            bf_g = self.get_base(qp_coors, diff=True, grad_axes=fc_bf_map[1])
            bf_g = nm.ascontiguousarray(bf_g[fc_bf_map[0]])
        else:
            bf_g = self.get_base(qp_coors, diff=True)

        if nm.allclose(bf_g, 0.0) and self.dim > 1:
            raise ValueError('zero basis function gradient!')

        if not is_face:
            ebf_g = poly_space.eval_base(qp_coors, diff=True, ori=ori,
                                         force_axis=True, transform=transform)
            size = ebf_g.nbytes * self.n_el
            site_config = Config()
            raise_if_too_large(size, site_config.refmap_memory_factor())
            se_conn, se_bf_bg = None, None
        else:
            se_conn, se_bf_bg, ebf_g = extra

        tdim = poly_space.geometry.dim
        if not is_face and tdim < self.dim:
            ecoors = tranform_coors_to_lower_dim(self.coors[self.conn, :],
                                                 tdim)
        else:
            ecoors = None

        margs = eval_mapping_data_in_qp(self.coors, self.conn,
                                        bf_g, weights, ebf_g, is_face=is_face,
                                        se_conn=se_conn, se_bf_bg=se_bf_bg,
                                        ecoors=ecoors)

        if bf is None:
            bf = nm.array([[[[0.]]]])
        elif len(bf.shape) == 3:
            bf = bf[None, ...]

        margs = (nm.ascontiguousarray(bf),) + margs + (tdim,)
        pycmap = PyCMapping(*margs)

        return pycmap



import os
import numpy as nm
import pytest

import sfepy.base.testing as tst

@pytest.fixture(scope='module')
def log_filename(output_dir):
    return os.path.join(output_dir, 'test_log.txt')

@pytest.fixture(scope='module')
def log(log_filename):
    from sfepy.base.log import Log

    log = Log([['x'], ['x^2', 'x^3']],
              plot_kwargs = [{},
                             [{'color' : 'b', 'ls' : '', 'marker' : 'o'},
                              {'color' : 'g', 'ls' : ':', 'marker' : 'x'}]],
              yscales=['log', 'linear'],
              xlabels=['x', 'x'], ylabels=['x', 'x^p'],
              is_plot=False,
              aggregate=0, sleep=0.0,
              log_filename=log_filename,
              formats=[['{:.3e}'], ['{:.5e}'] * 2])

    for x in nm.linspace(0, 1, 11):
        log(x, x**2, x**3, x=[x + 1, x])
        if nm.allclose(x, 0.5):
            log.plot_vlines([0], color='g', linewidth=2)
        if nm.allclose(x, 0.7):
            log.plot_vlines([1], color='g', linewidth=2)

    log(finished=True)
    return log

def test_log_rw(log_filename, log, output_dir):
    from sfepy.base.base import Output
    from sfepy.base.log import read_log, write_log

    log, info = read_log(log_filename)

    filename2 = os.path.join(output_dir, 'test_log2.txt')
    output = Output('', filename=filename2, quiet=True)

    write_log(output, log, info)

    log2, info2 = read_log(filename2)

    ok = True
    _ok = info == info2
    if not _ok:
        tst.report('headers are not equal!')
        tst.report(info)
        tst.report(info2)
    ok = ok and _ok

    for key, val2 in log2.items():
        val = log[key]
        _ok = nm.allclose(val[0], val2[0], rtol=0.0, atol=1e-14)
        if not _ok:
            tst.report('x values are not equal!')
            tst.report(val[0])
            tst.report(val2[0])
        ok = ok and _ok

        _ok = nm.allclose(val[1], val2[1], rtol=0.0, atol=1e-14)
        if not _ok:
            tst.report('y values are not equal!')
            tst.report(val[1])
            tst.report(val2[1])
        ok = ok and _ok

        _ok = nm.allclose(val[2], val2[2], rtol=0.0, atol=1e-14)
        if not _ok:
            tst.report('vlines are not equal!')
            tst.report(val[2])
            tst.report(val2[2])
        ok = ok and _ok

    assert ok



r"""
A script demonstrating the solution of the scalar Helmholtz equation for a
situation inspired by the physical problem of WiFi propagation in an apartment.

The example is an adaptation of the project found here:
https://bthierry.pages.math.cnrs.fr/course-fem/projet/2017-2018/

The boundary conditions are defined as perfect radiation conditions, meaning
that on the boundary waves will radiate outside.

The PDE for this physical process implies to find :math:`E(x, t)` for :math:`x
\in \Omega` such that:

.. math::
    \left\lbrace
    \begin{aligned}
      \Delta E+ k^2 n(x)^2 E = f_s && \forall x \in \Omega \\
      \partial_n E(x)-ikn(x)E(x)=0 && \forall x \text{ on } \partial \Omega
    \end{aligned}
    \right.


Usage
-----

The mesh of the appartement and the different material ids can be visualized
with the following::

    sfepy-view meshes/2d/helmholtz_apartment.vtk -e -2

The example is run from the top level directory as::

    sfepy-run sfepy/examples/acoustics/helmholtz_apartment.py

The result of the computation can be visualized as follows::

    sfepy-view helmholtz_apartment.vtk --color-map=seismic -f imag.E:wimag.E:f10%:p0 --camera-position="-5.14968,-7.27948,7.08783,-0.463265,-0.23358,-0.350532,0.160127,0.664287,0.730124"
"""
import numpy as nm
from sfepy import data_dir

f = 2.4 * 1e9  # change this to 2.4 or 5 if you want to simulate frequencies
               # typically used by your Wifi
c0 = 3e8  # m/s
k = 2 * nm.pi * f / c0

filename_mesh = data_dir + "/meshes/2d/helmholtz_apartment.vtk"

regions = {
    'Omega': 'all',
    'Walls': ('cells of group 1'),
    'Air': ('cells of group 2'),
    'Source': ('cells of group 3'),
    'Gamma': ('vertices of surface', 'facet'),
}

# air and walls have different material parameters, hence the 1. and 2.4 factors
materials = {
    'material': ({'kn_square': {
        'Air': (k * 1.) ** 2,
        'Walls': (k * 2.4) ** 2}, },),
    'boundary': ({'kn': 1j * k * 2.4},),
    'source_func': 'source_func',
}


def source_func(ts, coors, mode=None, **kwargs):
    """The source function for the antenna."""
    epsilon = 0.1
    x_center = nm.array([-3.23, -1.58])
    if mode == 'qp':
        dists = nm.abs(nm.sum(nm.square(coors - x_center), axis=1))
        val = 3 / nm.pi / epsilon ** 2 * (1 - dists / epsilon)
        return {'val': val[:, nm.newaxis, nm.newaxis]}


functions = {
    'source_func': (source_func,),
}

fields = {
    'electric_field': ('complex', 1, 'Omega', 1),
}

variables = {
    'E': ('unknown field', 'electric_field', 1),
    'q': ('test field', 'electric_field', 'E'),
}

ebcs = {
}

integrals = {
    'i': 2,
}

equations = {
    'Helmholtz equation':
        """- dw_laplace.i.Omega(q, E)
        + dw_dot.i.Omega(material.kn_square, q, E)
        + dw_dot.i.Gamma(boundary.kn, q, E)
        = dw_integrate.i.Source(source_func.val, q)
        """
}

solvers = {
    'ls': ('ls.auto_direct', {}),
    'newton': ('nls.newton', {
        'i_max': 1,
    }),
}

options = {
    'refinement_level': 3,
}



import os.path as op
import numpy as nm
import pytest

import sfepy
from sfepy.discrete import FieldVariable
from sfepy.discrete.fem import Mesh, FEDomain, Field
from sfepy.base.base import assert_
import sfepy.base.testing as tst

@pytest.fixture(scope='module')
def data():
    from sfepy.base.base import Struct

    mesh = Mesh.from_file('meshes/2d/square_unit_tri.mesh',
                          prefix_dir=sfepy.data_dir)
    domain = FEDomain('domain', mesh)

    omega = domain.create_region('Omega', 'all')

    field = Field.from_args('linear', nm.float64, 'scalar', omega,
                            approx_order=1)
    return Struct(omega=omega, field=field)

def test_mass_matrix(data):
    from sfepy.discrete.projections import create_mass_matrix

    field = data.field

    mtx = create_mass_matrix(field)

    assert_(mtx.shape == (field.n_nod, field.n_nod))
    assert_(abs(mtx.sum() - 1.0) < 1e-14)

def test_projection_tri_quad(data, output_dir):
    from sfepy.discrete.projections import make_l2_projection

    source = FieldVariable('us', 'unknown', data.field)

    coors = data.field.get_coor()
    vals = nm.sin(2.0 * nm.pi * coors[:,0] * coors[:,1])
    source.set_data(vals)

    name = op.join(output_dir,
                   'test_projection_tri_quad_source.vtk')
    source.save_as_mesh(name)

    mesh = Mesh.from_file('meshes/2d/square_quad.mesh',
                          prefix_dir=sfepy.data_dir)
    domain = FEDomain('domain', mesh)

    omega = domain.create_region('Omega', 'all')


    field = Field.from_args('bilinear', nm.float64, 'scalar', omega,
                            approx_order=1)

    target = FieldVariable('ut', 'unknown', field)

    make_l2_projection(target, source)

    name = op.join(output_dir,
                   'test_projection_tri_quad_target.vtk')
    target.save_as_mesh(name)

    bbox = data.field.domain.get_mesh_bounding_box()
    x = nm.linspace(bbox[0, 0] + 0.001, bbox[1, 0] - 0.001, 20)
    y = nm.linspace(bbox[0, 1] + 0.001, bbox[1, 1] - 0.001, 20)

    xx, yy = nm.meshgrid(x, y)
    test_coors = nm.c_[xx.ravel(), yy.ravel()].copy()

    vec1 = source.evaluate_at(test_coors)
    vec2 = target.evaluate_at(test_coors)

    ok = (nm.abs(vec1 - vec2) < 0.01).all()
    assert_(ok)

def test_projection_iga_fem():
    try:
        from igakit import igalib; igalib
    except ImportError:
        tst.report('iga-fem projection not-tested (missing igalib module)!')
        return

    from sfepy.discrete import FieldVariable
    from sfepy.discrete.fem import FEDomain, Field
    from sfepy.discrete.iga.domain import IGDomain
    from sfepy.mesh.mesh_generators import gen_block_mesh
    from sfepy.discrete.iga.domain_generators import gen_patch_block_domain
    from sfepy.discrete.projections import (make_l2_projection,
                                            make_l2_projection_data)

    shape = [10, 12, 12]
    dims = [5, 6, 6]
    centre = [0, 0, 0]
    degrees = [2, 2, 2]

    nurbs, bmesh, regions = gen_patch_block_domain(dims, shape, centre,
                                                   degrees,
                                                   cp_mode='greville',
                                                   name='iga')
    ig_domain = IGDomain('iga', nurbs, bmesh, regions=regions)

    ig_omega = ig_domain.create_region('Omega', 'all')
    ig_field = Field.from_args('iga', nm.float64, 1, ig_omega,
                               approx_order='iga', poly_space_base='iga')
    ig_u = FieldVariable('ig_u', 'parameter', ig_field,
                         primary_var_name='(set-to-None)')

    mesh = gen_block_mesh(dims, shape, centre, name='fem')
    fe_domain = FEDomain('fem', mesh)

    fe_omega = fe_domain.create_region('Omega', 'all')
    fe_field = Field.from_args('fem', nm.float64, 1, fe_omega,
                               approx_order=2)
    fe_u = FieldVariable('fe_u', 'parameter', fe_field,
                         primary_var_name='(set-to-None)')

    def _eval_data(ts, coors, mode, **kwargs):
        return nm.prod(coors**2, axis=1)[:, None, None]

    make_l2_projection_data(ig_u, _eval_data)

    make_l2_projection(fe_u, ig_u) # This calls ig_u.evaluate_at().

    coors = 0.5 * nm.random.rand(20, 3) * dims

    ig_vals = ig_u.evaluate_at(coors)
    fe_vals = fe_u.evaluate_at(coors)

    ok = nm.allclose(ig_vals, fe_vals, rtol=0.0, atol=1e-12)
    if not ok:
        tst.report('iga-fem projection failed!')
        tst.report('coors:')
        tst.report(coors)
        tst.report('iga fem diff:')
        tst.report(nm.c_[ig_vals, fe_vals, nm.abs(ig_vals - fe_vals)])

    assert_(ok)

def test_project_tensors(data):
    from sfepy.discrete import FieldVariable
    from sfepy.discrete.projections import project_by_component

    ok = True

    u = FieldVariable('u', 'parameter', data.field,
                      primary_var_name='(set-to-None)')
    u.set_constant(1.0)

    component = FieldVariable('component', 'parameter', data.field,
                              primary_var_name='(set-to-None)')

    nls_options = {'eps_a' : 1e-16, 'i_max' : 1}

    u_qp = u.evaluate()
    u2 = FieldVariable('u2', 'parameter', data.field,
                       primary_var_name='(set-to-None)')
    project_by_component(u2, u_qp, component, data.field.approx_order,
                         nls_options=nls_options)

    _ok = tst.compare_vectors(u(), u2())
    ok = ok and _ok

    gu_qp = u.evaluate(mode='grad')

    gfield = Field.from_args('gu', nm.float64, 2, data.field.region,
                             approx_order=data.field.approx_order)
    gu = FieldVariable('gu', 'parameter', gfield,
                       primary_var_name='(set-to-None)')

    project_by_component(gu, gu_qp, component, gfield.approx_order,
                         nls_options=nls_options)

    _ok = tst.compare_vectors(gu(), nm.zeros_like(gu()))
    ok = ok and _ok

    assert_(ok)



import numpy as nm

try:
    import sympy as sm
except ImportError:
    sm = None

import pytest

from sfepy.base.base import assert_, ordered_iteritems
import sfepy.base.testing as tst

def symarray(prefix, shape):
    """
    Copied from SymPy so that the tests pass for its different versions.
    """
    arr = nm.empty(shape, dtype=object)
    for index in nm.ndindex(shape):
        arr[index] = sm.Symbol('%s_%s' % (prefix, '_'.join(map(str, index))))
    return arr

def get_poly(order, dim, is_simplex=False):
    """
    Construct a polynomial of given `order` in space dimension `dim`,
    and integrate it symbolically over a rectangular or simplex domain
    for coordinates in [0, 1].
    """
    xs = symarray('x', dim)

    opd = max(1, int((order + 1) / dim))

    poly = 1.0
    oo = 0
    for ii, x in enumerate(xs):
        if ((oo + opd) > order) or (ii == (len(xs) - 1)):
            opd = max(order - oo, 0)

        poly *= (x**opd + 1)
        oo += opd

    assert_(oo == order)

    limits = [[xs[ii], 0, 1] for ii in range(dim)]
    if is_simplex:
        for ii in range(1, dim):
            for ip in range(0, ii):
                limits[ii][2] -= xs[ip]

    integral = sm.integrate(poly, *reversed(limits))

    return xs, poly, limits, integral

def test_weight_consistency():
    """
    Test if integral of 1 (= sum of weights) gives the domain volume.
    """
    from sfepy.discrete.quadratures import quadrature_tables

    ok = True
    for geometry, qps in ordered_iteritems(quadrature_tables):
        tst.report('geometry:', geometry)
        for order, qp in ordered_iteritems(qps):
            diff = nm.abs(qp.weights.sum() - qp.volume)
            _ok = diff < 1e-14
            tst.report('order %d: %s (%.2e)' % (order, _ok, diff))

            ok = ok and _ok

    assert_(ok)

def _test_quadratures(quad, geometry, order):
    dim = int(geometry[0])
    n_v = int(geometry[2])
    is_simplex = n_v == (dim + 1)

    porder = order if is_simplex else dim * order

    xs, poly, limits, integral = get_poly(porder, dim,
                                          is_simplex=is_simplex)

    tst.report('  polynomial:', poly)
    tst.report('  limits:', limits)
    tst.report('  integral:', integral)

    def fun(coors):
        vals = nm.empty(coors.shape[0], dtype=nm.float64)

        subs = {}
        for ir, cc in enumerate(coors):
            for ic, x in enumerate(xs):
                subs[x] = coors[ir,ic]
            vals[ir] = float(poly.subs(subs))

        return vals

    val = quad.integrate(fun, order=order, geometry=geometry)
    ok = nm.allclose(val, float(integral), rtol=0.0, atol=1e-11)

    tst.report('  sym. == num.: %f == %f -> %s' %
               (float(integral), val, ok))

    return ok, integral, val

@pytest.mark.slow
def test_quadratures():
    """
    Test if the quadratures have orders they claim to have, using
    symbolic integration by sympy.
    """
    from sfepy.discrete.quadratures import quadrature_tables, tp_geometries
    from sfepy.discrete.integrals import Integral

    if sm is None:
        tst.report('cannot import sympy, skipping')
        return

    quad = Integral('test_integral')

    ok = True
    failed = []
    for geometry, qps in ordered_iteritems(quadrature_tables):
        tst.report('geometry:', geometry)

        if geometry == '0_1':
            continue

        elif geometry in tp_geometries:
            iter_qp = range(1, 11)

        elif geometry == '2_3':
            iter_qp = range(1, 25)

        elif geometry == '3_4':
            iter_qp = range(1, 12)

        elif geometry == '3_6':
            iter_qp = range(4, 9)

        else:
            iter_qp = sorted(qps.keys())

        for order in iter_qp:
            tst.report('order:', order)

            _ok, integral, val = _test_quadratures(quad, geometry, order)

            if not _ok:
                failed.append((geometry, order, float(integral), val))

            ok = ok and _ok

    if not ok:
        tst.report('failed:')
        for aux in failed:
            tst.report(aux, '%.1e' % abs(aux[2] - aux[3]))

    assert_(ok)



"""
Computational domain, consisting of the mesh and regions.
"""
from __future__ import absolute_import
import numpy as nm

from sfepy.base.base import output, Struct
from .geometry_element import GeometryElement
from sfepy.discrete import Domain, PolySpace
from sfepy.discrete.fem.refine import refine_2_3, refine_2_4, refine_3_4, \
    refine_3_8, refine_1_2
from sfepy.discrete.fem.fe_surface import FESurface


class FEDomain(Domain):
    """
    Domain is divided into groups, whose purpose is to have homogeneous
    data shapes.
    """

    def __init__(self, name, mesh, verbose=False, **kwargs):
        """Create a Domain.

        Parameters
        ----------
        name : str
            Object name.
        mesh : Mesh
            A mesh defining the domain.
        """
        Domain.__init__(self, name, mesh=mesh, verbose=verbose, **kwargs)

        self.geom_els = geom_els = {}
        for ig, desc in enumerate(mesh.descs):
            gel = GeometryElement(desc)

            if gel.dim > 0:
                # Create geometry elements of dimension - 1.
                gel.create_surface_facet()

            geom_els[desc] = gel

        for gel in geom_els.values():
            key = gel.get_interpolation_name()

            gel.poly_space = PolySpace.any_from_args(key, gel, 1)
            gel = gel.surface_facet
            if isinstance(gel, dict):
                for v in gel.values():
                    key = v.get_interpolation_name()
                    v.poly_space = PolySpace.any_from_args(key, v, 1)
            elif gel is not None:
                key = gel.get_interpolation_name()
                gel.poly_space = PolySpace.any_from_args(key, gel, 1)

        self.vertex_set_bcs = self.mesh.nodal_bcs
        self.cmesh_tdim = self.mesh.cmesh_tdim

        # Must be before creating derived connectivities.
        self.fix_element_orientation()

        from sfepy.discrete.fem.geometry_element import create_geometry_elements
        gels = create_geometry_elements()
        max_tdim = 0
        for cmesh in self.mesh.cmesh_tdim:
            if cmesh is not None:
                cmesh.set_local_entities(gels)
                cmesh.setup_entities()
                max_tdim = max(max_tdim, cmesh.tdim)

        self.cmesh = self.cmesh_tdim[max_tdim]
        n_nod, dim = self.mesh.coors.shape
        self.shape = Struct(n_nod=n_nod, dim=dim, tdim=self.cmesh.tdim,
                            n_el=self.cmesh.n_el,
                            n_gr=len(self.geom_els))

        self.reset_regions()
        self.clear_surface_groups()

    def get_mesh_coors(self, actual=False):
        """
        Return the coordinates of the underlying mesh vertices.
        """
        if actual and hasattr(self.mesh, 'coors_act'):
            return self.mesh.coors_act
        else:
            return self.mesh.coors

    def get_mesh_bounding_box(self):
        """
        Return the bounding box of the underlying mesh.

        Returns
        -------
        bbox : ndarray (2, dim)
            The bounding box with min. values in the first row and max. values
            in the second row.
        """
        return self.mesh.get_bounding_box()

    def get_diameter(self):
        """
        Return the diameter of the domain.

        Notes
        -----
        The diameter corresponds to the Friedrichs constant.
        """
        bbox = self.get_mesh_bounding_box()
        return (bbox[1,:] - bbox[0,:]).max()

    def fix_element_orientation(self, geom_els=None, force_check=False):
        """
        Ensure element vertices ordering giving positive cell volumes.
        """
        from sfepy.discrete.common.extmods.cmesh import orient_elements

        if geom_els is None:
            geom_els = self.geom_els

        for key, gel in geom_els.items():
            ori = gel.orientation
            cmesh = self.cmesh_tdim[gel.dim]
            if (cmesh.tdim != cmesh.dim) and (not force_check):
                output('warning: mesh with topological dimension %d lower than'
                       ' space dimension %d' % (cmesh.tdim, cmesh.dim))
                output('- element orientation not checked!')
                return

            cells = nm.where(cmesh.cell_types == cmesh.key_to_index[gel.name])
            cells = cells[0].astype(nm.uint32)

            itry = 0
            while itry < 2:
                flag = -nm.ones(cmesh.n_el, dtype=nm.int32)

                # Changes orientation if it is wrong according to swap*!
                # Changes are indicated by positive flag.
                orient_elements(flag, cmesh, cells, gel.dim,
                                ori.roots, ori.vecs,
                                ori.swap_from, ori.swap_to)

                flag = flag[cells]

                if nm.all(flag == 0):
                    if itry > 0: output('...corrected')
                    itry = -1
                    break

                output('warning: bad element orientation, trying to correct...')
                itry += 1

            if itry == 2 and flag[0] != -1:
                raise RuntimeError('elements cannot be oriented! (%s)' % key)

            elif flag[0] == -1:
                output('warning: element orienation not checked')

    def get_conn(self, ret_gel=False, tdim=None, cells=None):
        """
        Get the cell-vertex connectivity and, if `ret_gel` is True, also the
        corresponding reference geometry element. If `tdim` is not None get
        the connectivity of the cells with topological dimension `tdim`.
        """
        cmesh = self.cmesh if tdim is None else self.cmesh_tdim[tdim]
        if cells is None:
            conn = cmesh.get_conn(cmesh.tdim, 0).indices
            conn = conn.reshape((cmesh.n_el, -1)).astype(nm.int32)
        else:
            conn = cmesh.get_conn(cmesh.tdim, 0)
            aux = conn.offsets[cells]
            nv = set(conn.offsets[cells + 1] - aux)
            if len(nv) != 1:
                raise ValueError('different cell types in a region!')

            nc = cells.shape[0]
            nv = nv.pop()
            idxs = (nm.repeat(conn.offsets[cells], nv) +
                    nm.tile(nm.arange(nv), nc))
            conn = conn.indices[idxs].reshape((nc, nv))

        name = f'{cmesh.tdim}_{conn.shape[1]}'
        if ret_gel:
            gel = None
            for gv in self.geom_els.values():
                if gv.name == name:
                    gel = gv
                    break

            return conn, gel

        else:
            return conn

    def get_element_diameters(self, cells, volume, mode, square=True):
        conn, gel = self.get_conn(ret_gel=True)
        coors = self.get_mesh_coors()
        lconn = conn[cells]
        ed = gel.edges

        if mode == 0 or mode == 2:
            vv = coors[lconn[:, ed[:, 1]]] - coors[lconn[:, ed[:, 0]]]
            v0 = nm.max(nm.sum(vv**2, axis=2), axis=1)
            out = v0

        if mode == 1 or mode == 2:
            v1 = nm.power(0.16 * volume, 1.0 / gel.dim).squeeze()
            out = v1

        if mode == 2:
            out = nm.max(nm.stack([v0, v1]), axis=0)

        return out if square else nm.sqrt(out)

    def clear_surface_groups(self):
        """
        Remove surface group data.
        """
        self.surface_groups = {}

    def create_surface_group(self, region):
        """
        Create a new surface group corresponding to `region` if it does
        not exist yet.

        Notes
        -----
        Surface groups define surface facet connectivity that is needed
        for :class:`sfepy.discrete.fem.mappings.FEMapping`.
        """
        groups = self.surface_groups
        if region.name not in groups:
            name = 'surface_group_%s' % (region.name)
            surface_group = FESurface.from_region(name, region)

            groups[region.name] = surface_group

    def refine(self):
        """
        Uniformly refine the domain mesh.

        Returns
        -------
        domain : FEDomain instance
            The new domain with the refined mesh.

        Notes
        -----
        Works only for meshes with single element type! Does not
        preserve node groups!
        """
        if len(self.geom_els) != 1:
            msg = 'refine() works only for meshes with single element type!'
            raise NotImplementedError(msg)

        el_type = list(self.geom_els.values())[0].name
        if el_type == '1_2':
            mesh = refine_1_2(self.mesh)

        elif el_type == '2_3':
            mesh = refine_2_3(self.mesh)

        elif el_type == '2_4':
            mesh = refine_2_4(self.mesh)

        elif el_type == '3_4':
            mesh = refine_3_4(self.mesh)

        elif el_type == '3_8':
            mesh = refine_3_8(self.mesh)

        else:
            msg = 'unsupported element type! (%s)' % el_type
            raise NotImplementedError(msg)

        domain = FEDomain(self.name + '_r', mesh)

        return domain
